import json
from typing import Any
from pydantic import Field, model_validator

from gwenflow.tools import BaseTool


WIKIPEDIA_MAX_QUERY_LENGTH = 300


class WikipediaBaseTool(BaseTool):

    client: Any
    lang: str = "en"
    top_k_results: int = 5
    doc_content_chars_max: int = 4000

    @model_validator(mode="before")
    @classmethod
    def validate_environment(cls, values: Any) -> Any:
        """Validate that the python package exists in environment."""
        try:
            import wikipedia
            lang = values.get("lang", "en")
            wikipedia.set_lang(lang)
            values["client"] = wikipedia
        except ImportError:
            raise ImportError(
                "Could not import wikipedia python package. "
                "Please install it with `pip install wikipedia`."
            )
        return values


class WikipediaTool(WikipediaBaseTool):

    name: str = "WikipediaTool"
    description: str = (
        "A wrapper around Wikipedia. "
        "Useful for when you need to answer general questions about "
        "people, places, companies, facts, historical events, or other subjects. "
        "Input should be a search query."
    )

    def _run(self, query: str = Field(description="query to look up on wikipedia")) -> str:

        documents = []

        doc_content_chars = 0
    
        page_titles = self.client.search(query[:WIKIPEDIA_MAX_QUERY_LENGTH], results=self.top_k_results)
        for page_title in page_titles[:self.top_k_results]:
            try:
                page = self.client.page(title=page_title, auto_suggest=False)
                summary = {"title": page_title, "summary": page.summary}
                if (doc_content_chars+len(page_title)+len(page.summary)+20) < self.doc_content_chars_max:
                    documents.append(summary)
                    doc_content_chars += len(page_title) + len(page.summary) + 20
            except Exception as e:
                pass

        if not documents:
            return "No good Wikipedia Search Result was found"

        return json.dumps([doc for doc in documents])



PROMPT_STEPS = """Before providing your final answer, follow these steps:

1. List the key topics or concepts you've identified from the task.
2. For each topic, list potential information you plan to search.
3. For each information:
   a. Note the most relevant information you find.
   b. Include any important quotes, with proper citation.
   c. Highlight any statistical data or key facts.
4. Identify any contradictions or gaps in the information you've found.
5. Summarize your key findings and how they relate to the task.

This structured approach will help you organize your thoughts and ensure a thorough response.
"""


PROMPT_TASK = """
## You have received the following task from your manager:

<task>
{task}
</task>
"""

PROMPT_TOOLS = """
## Tools
Your objective is to thoroughly research your task using the following tools as your primary source and provide a detailed and informative answer.

You have access to the following tools:
<tools>
{tools}
</tools>
"""

PROMPT_TOOLS_REACT_GUIDELINES = """
The tool call you write is an 'Action'. After the tool is executed, you will get the result of the tool call as an 'Observation'.
This Action/Observation can repeat N times. You should take several steps when needed.
You can use the result of the previous 'Action' as input for the next 'Action'.
The 'Observation' will always be a string. Then you can use it as input for the next 'Action'.
"""



import logging
import hashlib
import json
from typing import List, Optional, Dict, Any

try:
    import lancedb
    from lancedb.pydantic import Vector, LanceModel
except ImportError:
    raise ImportError("`lancedb` not installed.")

try:
    import pyarrow as pa
except ImportError:
    raise ImportError("`pyarrow` not installed.")


from gwenflow.vector_stores.base import VectorStoreBase
from gwenflow.embeddings import Embeddings, GwenlakeEmbeddings
from gwenflow.reranker import Reranker
from gwenflow.types import Document


logger = logging.getLogger(__name__)


class LanceDB(VectorStoreBase):

    def __init__(
        self,
        uri: lancedb.URI,
        collection: str = "default",
        client: Optional[lancedb.DBConnection] = None,
        embeddings: Embeddings = GwenlakeEmbeddings(),
        reranker: Optional[Reranker] = None,
        api_key: str = None,
    ):

        # Embedder
        self.embeddings = embeddings

        # reranker
        self.reranker = reranker

        # collection and uri
        self.collection = collection
        self.uri: lancedb.URI = uri
        self.table: lancedb.db.LanceTable = None

        self.client: lancedb.DBConnection = client or lancedb.connect(uri=self.uri, api_key=api_key)
        self.create()

    def create(self):
        if not self.exists():

            class schema(LanceModel):
                id: str
                vector: Vector(self.embeddings.dimensions)
                payload: str

            logger.debug(f"Creating collection: {self.collection}")
            self.table = self.client.create_table(self.collection, schema=schema, mode="overwrite", exist_ok=True)
            # self.table.create_index(column='vector', index_type='IVF_PQ', metric="cosine", num_partitions=256, num_sub_vectors=32)

        elif self.table is None:
            self.table = self.client.open_table(self.collection)


    def exists(self) -> bool:
        if self.client:
            if self.collection in self.client.table_names():
                return True
        return False
    
    def get_collections(self) -> list:
        if self.client:
            return self.client.table_names()
        return []

    def insert(self, documents: list[Document]):
        logger.info(f"Inserting {len(documents)} documents into collection {self.collection}")

        data = []
        for document in documents:
            text_for_id = document.id
            if text_for_id is None:
                text_for_id = document.content
            _id = hashlib.md5(text_for_id.encode(), usedforsecurity=False).hexdigest()
            _embeddings = self.embeddings.embed_documents([document.content])[0]
            _payload = document.metadata
            _payload["content"] = document.content
            data.append(
                dict(
                    id=_id,
                    vector=_embeddings,
                    payload=json.dumps(_payload),
                )
            )
    
        if len(data) > 0:
            for d in data:
                self.table.delete(f"id='{ d['id'] }'")
            self.table.add(data)


    def search(self, query: str, limit: int = 5, filters: dict = None) -> list[Document]:

        query_embedding = self.embeddings.embed_query(query)
        if query_embedding is None:
            logger.error(f"Error getting embedding for Query: {query}")
            return []

        results = self.table.search(
            query=query_embedding,
            vector_column_name="vector",
        ).metric("cosine").limit(limit).to_list()

        documents = []
        for item in results:
            payload = json.loads(item["payload"])
            documents.append(
                Document(
                    id=item["id"],
                    content=payload.pop("content"),
                    metadata=payload,
                    score=1-item["_distance"],
                )
            )
    
        if self.reranker:
            documents = self.reranker.rerank(query=query, documents=documents)

        return documents


    def drop(self):
        if self.exists():
            logger.debug(f"Deleting collection: {self.collection}")
            self.client.drop_table(self.collection)

    def count(self) -> int:
        if self.exists():
            return self.table.count_rows()
        return 0

    def info(self) -> dict:
        return {}
    
    def delete(self, id: int):
        return False

    def get(self, id: int) -> dict:        
        return None

    def list(self, filters: dict = None, limit: int = 100) -> list:
        return []



from gwenflow.llms.openai import ChatOpenAI
from gwenflow.llms.azure_openai import ChatAzureOpenAI
from gwenflow.llms.gwenlake import ChatGwenlake
from gwenflow.llms.ollama import ChatOllama
from gwenflow.llms.deepseek import ChatDeepSeek

__all__ = [
    "ChatOpenAI",
    "ChatAzureOpenAI",
    "ChatGwenlake",
    "ChatOllama",
    "ChatDeepSeek",
]


import inspect

def function_to_json(func, name: str = None, description: str = None) -> dict:
    """
    Converts a Python function into a JSON-serializable dictionary
    that describes the function's signature, including its name,
    description, and parameters.

    Args:
        func: The function to be converted.

    Returns:
        A dictionary representing the function's signature in JSON format.
    """
    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
        type(None): "null",
    }

    try:
        signature = inspect.signature(func)
    except ValueError as e:
        raise ValueError(
            f"Failed to get signature for function {func.__name__}: {str(e)}"
        )

    parameters = {}
    for param in signature.parameters.values():
        try:
            param_type = type_map.get(param.annotation, "string")
        except KeyError as e:
            raise KeyError(
                f"Unknown type annotation {param.annotation} for parameter {param.name}: {str(e)}"
            )

        if hasattr(param.default, "description"):
            parameters[param.name] = {
                "description": param.default.description,
                "type": param_type,
            }
        else:
            parameters[param.name] = {"type": param_type}

    required = [
        param.name
        for param in signature.parameters.values()
        if param.default == inspect._empty
    ]

    return {
        "type": "function",
        "function": {
            "name": name or func.__name__,
            "description": description or func.__doc__ or "",
            "parameters": {
                "type": "object",
                "properties": parameters,
                "required": required,
            },
        },
    }



import json
from pydantic import Field

from gwenflow.tools import BaseTool
from gwenflow.readers.pdf import PDFReader


class PDFTool(BaseTool):

    name: str = "PDFTool"
    description: str = "This function reads a PDF and returns its content."

    def _run(self, file: str = Field(description="The path of the PDf file to read.")) -> str:
        reader = PDFReader()
        documents = reader.read(file)
        return json.dumps([doc.to_dict() for doc in documents])




import io
from typing import List, Union
from pathlib import Path

from gwenflow.types import Document
from gwenflow.readers.base import Reader
from gwenflow.utils import logger


class PDFReader(Reader):

    def read(self, file: Union[Path, io.BytesIO]) -> List[Document]:

        try:
            import pymupdf
        except ImportError:
            raise ImportError("PyMuPDF is not installed. Please install it with `pip install PyMuPDF`.")

        try:
            filename = self.get_file_name(file)
            content  = self.get_file_content(file)

            documents = []
            for page in pymupdf.open(stream=content, filetype="pdf"):
                text = page.get_text()
                safe_text = text.encode('utf-8', errors='ignore').decode('utf-8')
                tables = []
                for table in page.find_tables():
                    tables.append(table.extract())
                metadata = dict(filename=filename, page=page.number+1, tables=tables, images=[])
                doc = Document(id=self.key(f"{filename}_{page.number+1}"), content=safe_text, metadata=metadata)
                documents.append(doc)

        except Exception as e:
            logger.error(f"Error reading file: {e}")
            return []

        return documents




from abc import ABC
from typing import cast
from pathlib import Path, PurePosixPath

import fsspec
from fsspec.implementations.local import LocalFileSystem
from tqdm import tqdm

from gwenflow.types import Document
from gwenflow.readers.json import JSONReader
from gwenflow.readers.pdf import PDFReader

default_file_reader_cls = {
    ".json": JSONReader,
    ".pdf": PDFReader,
}


def get_default_fs() -> fsspec.AbstractFileSystem:
    return LocalFileSystem()


class SimpleDirectoryReader(ABC):
    """SimpleDirectoryReader"""

    def __init__(
        self,
        input_dir: str | None = None,
        input_files: list | None = None,
        exclude: list | None = None,
        exclude_hidden: bool = True,
        errors: str = "ignore",
        recursive: bool = True,
        encoding: str = "utf-8",
        filename_as_id: bool = False,
        required_exts: list[str] | None = None,
        raise_on_error: bool = False,
    ) -> None:

        if not input_dir and not input_files:
            raise ValueError("Must provide either `input_dir` or `input_files`.")

        self.fs = get_default_fs()

        self.exclude = exclude
        self.exclude_hidden = exclude_hidden
        self.errors = errors
        self.recursive = recursive
        self.encoding = encoding
        self.filename_as_id = filename_as_id
        self.required_exts = required_exts
        self.raise_on_error = raise_on_error

        if input_files:
            self.input_files = []
            for path in input_files:
                if not self.fs.isfile(path):
                    raise ValueError(f"File {path} does not exist.")
                input_file = Path(path)
                self.input_files.append(input_file)

        elif input_dir:
            if not self.fs.isdir(input_dir):
                raise ValueError(f"Directory {input_dir} does not exist.")
            self.input_dir = Path(input_dir)
            self.exclude = exclude
            self.input_files = self._add_files(self.input_dir)

    @staticmethod
    def is_hidden(path: Path | PurePosixPath) -> bool:
        return any(
            part.startswith(".") and part not in [".", ".."] for part in path.parts
        )

    def _add_files(self, input_dir: Path | PurePosixPath) -> list[Path | PurePosixPath]:

        all_files = []

        if self.recursive:
            list_files = input_dir.rglob("*")
        else:
            list_files = input_dir.glob("*")
        
        # only keep required_exts
        for file in list_files:
            if self.required_exts:
                if file.suffix.lower() in self.required_exts:
                    all_files.append(file)
            else:
                all_files.append(file)

        return list(set(all_files)) # remove duplicates

    
    def read(self, show_progress: bool = False) -> list[Document]:

        documents = []

        files_to_process = self.input_files

        if show_progress:
            files_to_process = tqdm(self.input_files, desc="Reading files", unit="file")

        for input_file in files_to_process:
            if SimpleDirectoryReader.is_hidden(input_file):
                continue
            documents.extend(
                SimpleDirectoryReader.read_file(
                    input_file=input_file,
                    filename_as_id=self.filename_as_id,
                    encoding=self.encoding,
                    errors=self.errors,
                    raise_on_error=self.raise_on_error,
                )
            )
        return documents
    
    @staticmethod
    def read_file(
        input_file: Path | PurePosixPath,
        filename_as_id: bool = False,
        encoding: str = "utf-8",
        errors: str = "ignore",
        raise_on_error: bool = False,
    ) -> list[Document]:
        
        documents = []

        file_suffix = input_file.suffix.lower()
        if not file_suffix:
            return []
        
        # supported formats
        supported_suffix = list(default_file_reader_cls.keys())
        if file_suffix in supported_suffix:
            try:
                reader = default_file_reader_cls[file_suffix]()
                documents = reader.read(file=input_file)
            except Exception as e:
                if raise_on_error:
                    raise Exception("Error loading file") from e
                print(f"Failed to load file {input_file} with error: {e}. Skipping...", flush=True)
                return []
        
        # other formats
        else:
            fs = get_default_fs()
            with fs.open(input_file, errors=errors, encoding=encoding) as f:
                content = cast(bytes, f.read()).decode(encoding, errors=errors)

            doc = Document(content=content, name=str(input_file))  # type: ignore

            if filename_as_id:
                doc.id = str(input_file)

            documents.append(doc)

        return documents


from gwenflow.utils.logger import (
    logger,
    set_log_level_to_debug,
    set_log_level_to_info,
)
from gwenflow.utils.bytes import bytes_to_b64_str
from gwenflow.utils.tokens import (
    num_tokens_from_string,
    num_tokens_from_messages,
)
from gwenflow.utils.json import to_json, extract_json_str

__all__ = [
    "logger",
    "set_log_level_to_debug",
    "set_log_level_to_info",
    "bytes_to_b64_str",
    "num_tokens_from_string",
    "num_tokens_from_messages",
    "to_json",
    "extract_json_str",
]


from abc import ABC, abstractmethod
from pydantic import BaseModel
from typing import Optional

from gwenflow.types import Document


class VectorStoreBase(ABC):

    @abstractmethod
    def get_collections(self):
        """List all collections."""
        pass

    @abstractmethod
    def create(self):
        """Create collection."""
        pass

    @abstractmethod
    def drop(self):
        """Delete collection."""
        pass

    @abstractmethod
    def count(self) -> int:
        """Count points in collection."""
        pass

    @abstractmethod
    def info(self):
        """Get information about collection."""
        pass

    @abstractmethod
    def insert(self, documents: list[Document]):
        """Insert documents into collection."""
        pass

    @abstractmethod
    def search(self, query, limit=5, filters=None) -> list[Document]:
        """Search for similar vectors."""
        pass

    @abstractmethod
    def delete(self, id):
        """Delete a vector by ID."""
        pass

    @abstractmethod
    def get(self, id):
        """Retrieve a vector by ID."""
        pass



import json

try:
    import boto3
except ImportError:
    raise ImportError("Please install boto3 with `pip install boto3`.")


def aws_get_instance_ip_address(instance_id, region_name="eu-west-3"):
    ec2 = boto3.resource('ec2', region_name=region_name)
    instance = ec2.Instance(instance_id)
    return instance.private_ip_address

def aws_s3_list_files(bucket: str, prefix: str, recursive: bool = True):
    files = []
    client = boto3.client('s3')
    if recursive:
        for key in client.list_objects(Bucket=bucket, Prefix=prefix)['Contents']:
            files.append(key['Key'])
    else:
        for key in client.list_objects(Bucket=bucket, Prefix=prefix, Delimiter='/')['CommonPrefixes']:
            files.append(key['Prefix'])
    return files

def aws_s3_read_file(bucket: str, file: str):
    client = boto3.client('s3')
    response = client.get_object(Bucket=bucket, Key=file)
    return response['Body'].read()

def aws_s3_delete_folder(bucket: str, prefix: str):
    if len(prefix)<35: # protect repo if prefix is not uuid4
        return False
    client = boto3.client('s3')
    for key in client.list_objects(Bucket=bucket, Prefix=prefix):
        key.delete()
    return True

def aws_s3_read_json_file(bucket: str, file: str):
    data = aws_s3_read_file(bucket=bucket, file=file)
    return json.loads(data.decode('utf-8'))

def aws_s3_read_text_file(bucket: str, file: str):
    data = aws_s3_read_file(bucket=bucket, file=file)
    return data.decode('utf-8')

def aws_s3_uri_to_bucket_key(uri: str):
    try:
        tmp = uri.replace("s3://", "")
        bucket = tmp.split("/")[0]
        key = tmp.replace(bucket, "").strip("/")
        return bucket, key
    except:
        pass
    return None, None




from typing import Optional, Dict
from pydantic import BaseModel


class ActionReasoningStep(BaseModel):
    thought: str
    action: Optional[str] = None
    action_input: Optional[Dict] = None
    response : Optional[str] = None
    is_done: bool = False



from gwenflow.types.chat_message import ChatMessage, messages_to_dict, messages_to_openai
from gwenflow.types.chat_completion import ChatCompletion, ChatCompletionMessage, ChatCompletionMessageToolCall, Function
from gwenflow.types.chat_completion_chunk import ChatCompletionChunk
from gwenflow.types.document import Document


__all__ = [
    "messages_to_dict",
    "messages_to_openai",
    "ChatMessage",
    "ChatCompletion",
    "ChatCompletionChunk",
    "ChatCompletionMessage",
    "ChatCompletionMessageToolCall",
    "Document",
    "Function",
]



from typing import List, Dict, Any, Optional
from pydantic import BaseModel, model_validator, field_validator, Field, ConfigDict

import hashlib
import uuid
from langchain_text_splitters import RecursiveCharacterTextSplitter

from gwenflow.types.document import Document
from gwenflow.embeddings import Embeddings, GwenlakeEmbeddings
from gwenflow.vector_stores.base import VectorStoreBase
from gwenflow.vector_stores.qdrant import Qdrant
from gwenflow.utils import logger


MIN_CONTENT_LENGTH = 20


class Knowledge(BaseModel):

    vector_db: Optional[VectorStoreBase] = Field(None, validate_default=True)
    chunk_size: Optional[int] = 500
    chunk_overlap: Optional[int] = 0

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    
    @field_validator("vector_db", mode="before")
    @classmethod
    def set_vector_db(cls, v: Optional[VectorStoreBase]) -> VectorStoreBase:
        if not v:
            try:
                collection_name = uuid.uuid4().hex
                v = Qdrant(collection=collection_name, embeddings=GwenlakeEmbeddings(model="multilingual-e5-base"), on_disk=False)
            except Exception as e:
                logger.error(f"Error creating knowledge: {e}")
        return v
    
    def search(self, query: str, limit: int = 5, filters: dict = None) -> list[Document]:
        try:
            if not self.vector_db:
                return []
            return self.vector_db.search(query, limit=limit, filters=filters)
        except Exception as e:
            logger.error(f"Error searching for documents: {e}")
        return []

    def _unique_key(self, text: str):
        return hashlib.md5(text.encode(), usedforsecurity=False).hexdigest()
    
    def load_document(self, document: Document) -> bool:
        if not self.vector_db:
            return False

        try:
            docs = []
            text_splitter = RecursiveCharacterTextSplitter(chunk_size=self.chunk_size, chunk_overlap=self.chunk_overlap)

            for doc in text_splitter.create_documents([document.content]):
                if len(doc.page_content) > MIN_CONTENT_LENGTH:
                    docs.append(
                        Document(
                            id=self._unique_key(doc.page_content), # id based on content to remove content duplicates in knowledge
                            content=doc.page_content,
                        )
                    )

            if len(docs)>0:
                self.vector_db.insert(docs)

            return True
        
        except Exception as e:
            logger.error(f"Error loading document: {e}")

        return False
    
    def load_documents(self, documents: List[Any]) -> bool:
        for document in documents:
            if isinstance(document, str):
                document = Document(content=document)
            self.load_document(document)
    


from __future__ import annotations

from pydantic import BaseModel, model_validator
from pathlib import Path
from typing import Any, Optional, Union
from string import Formatter


def _get_template_variables(template: str) -> list[str]:
    """Get the variables from the template.

    Args:
        template: The template string.
        template_format: The template format. Should be one of "f-string" or "jinja2".

    Returns:
        The variables from the template.

    Raises:
        ValueError: If the template format is not supported.
    """
    input_variables = {
        v for _, v, _, _ in Formatter().parse(template) if v is not None
    }

    return sorted(input_variables)


class PromptTemplate(BaseModel):
    """Prompt template for a language model."""

    template: str
    """The prompt template."""

    input_variables: list[str]
    """The prompt input variables."""
    
    @model_validator(mode="before")
    @classmethod
    def pre_init_validation(cls, values: dict) -> Any:
        """Check that template and input variables are consistent."""
        if values.get("template") is None:
            # pydantic will fail if template is not provided.
            return values
        values["input_variables"] = _get_template_variables(values["template"])
        return values

    def __str__(self) -> dict:
        return self.template

    def format(self, **kwargs: Any) -> str:
        """Format the prompt with the inputs.

        Args:
            kwargs: Any arguments to be passed to the prompt template.

        Returns:
            A formatted string.
        """
        return self.template.format(**kwargs)

    @classmethod
    def from_file(
        cls,
        template_file: Union[str, Path],
        encoding: Optional[str] = None,
    ) -> PromptTemplate:
        """Load a prompt from a file.

        Args:
            template_file: The path to the file containing the prompt template.
            encoding: The encoding system for opening the template file.
                If not provided, will use the OS default.

        Returns:
            The prompt loaded from the file.
        """
        with open(str(template_file), encoding=encoding) as f:
            template = f.read()
        return cls.from_template(template=template, role="user")

    @classmethod
    def from_template(
        cls,
        template: str,
        role: str = "user",
    ) -> PromptTemplate:
        """Load a prompt template from a template.

        Args:
            template: The template to load.
            kwargs: Any other arguments to pass to the prompt template.

        Returns:
            The prompt template loaded from the template.
        """
        return cls(template=template, role=role)


from pydantic import BaseModel, ConfigDict, Field, SecretStr, model_validator
from typing import Any, Dict, List, Optional, cast
import os
import re
import requests

from gwenflow.embeddings.base import Embeddings

EMBEDDING_DIMS = {
    "e5-base-v2": 768,
    "e5-large-v2": 1024,
    "multilingual-e5-base": 768,
    "multilingual-e5-large": 1024,
}

EMBEDDING_WITH_PASSAGE = list(EMBEDDING_DIMS.keys())


class GwenlakeEmbeddings(Embeddings):
    """Gwenlake embedding models."""

    api_base: str = "https://api.gwenlake.com/v1/embeddings"
    api_key: Optional[SecretStr] = None
    
    @model_validator(mode="before")
    @classmethod
    def validate_environment(cls, values: Dict) -> Dict:
        values["api_key"] = os.getenv("GWENLAKE_API_KEY")
        if "model" not in values:
            values["model"] = "e5-base-v2"
        values["dimensions"] = EMBEDDING_DIMS[ values["model"] ]
        return values

    def _embed(self, input: List[str]) -> List[List[float]]:

        api_key = cast(SecretStr, self.api_key).get_secret_value()

        payload = {"input": input, "model": self.model}

        # HTTP headers for authorization
        headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
        }

        # send request
        try:
            response = requests.post(self.api_base, headers=headers, json=payload)
        except requests.exceptions.RequestException as e:
            raise ValueError(f"Error raised by inference endpoint: {e}")
        
        if response.status_code != 200:
            raise ValueError(
                f"Error raised by inference API: rate limit exceeded.\nResponse: "
                f"{response.text}"
            )

        parsed_response = response.json()
        if "data" not in parsed_response:
            raise ValueError("Error raised by inference API.")

        embeddings = []
        for e in parsed_response["data"]:
            embeddings.append(e["embedding"])
        
        return embeddings

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Call out to Gwenlake's embedding endpoint.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """
        batch_size = 100
        embeddings = []
        try:
            for i in range(0, len(texts), batch_size):
                i_end = min(len(texts), i+batch_size)
                batch = texts[i:i_end]
                batch_processed = []
                for text in batch:
                    text = text.replace("\n", " ")
                    text = re.sub(' +', ' ', text)
                    text = text.strip()
                    if self.model in EMBEDDING_WITH_PASSAGE and not text.startswith("passage: "):
                        text = "passage: " + text
                    batch_processed.append(text)
                embeddings += self._embed(batch_processed)
        except:
            return None
        return embeddings


    def embed_query(self, text: str) -> List[float]:
        """Call out to Gwenlake's embedding endpoint.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        text = text.replace("\n", " ")
        text = re.sub(' +', ' ', text)
        text = text.strip()
        if self.model in EMBEDDING_WITH_PASSAGE and not text.startswith("query: "):
            text = "query: " + text
        return self._embed([text])[0]


PROMPT_REACT = """
## Output Format

If you need to use a tool, please answer using the following format:

```
Thought: you should always think about what to do.
Action: tool name (one of {tool_names}) if using a tool.
Action Input: the input to the tool, in a JSON format representing the kwargs (e.g. {{"input": "hello world", "num_beams": 5}})
```

Please ALWAYS start with a Thought.

If some instructions are provided under the tags <thinking></thinking>, please follow these instructions.

Please use a valid JSON format for the Action Input. Do NOT do this {{'input': 'hello world', 'num_beams': 5}}.

If this format is used, the user will respond in the following format:

```
Observation: tool response
```

You should keep repeating the above format until you have enough information to answer the question without using any more tools. At that point, you MUST respond in the following two format:

```
Thought: the final thought to the original input question
Final Answer: the final answer to the original input question
```

## Remember:
- Please use a valid JSON format for the Action Input. Do NOT do this {{'input': 'hello world', 'num_beams': 5}}.
- You should keep repeating the above format until you have enough information to answer the question without using any more tools.
"""




from typing import List, Dict, Any, Optional
from pydantic import BaseModel, model_validator, field_validator, Field

import yaml

from gwenflow.agents import Agent
from gwenflow.tools import Tool, BaseTool
from gwenflow.utils import logger


MAX_TRIALS=5


class Flow(BaseModel):

    agents: List[Agent] = []
    manager: Optional[Agent] = None
    llm: Any = None
    tools: List[BaseTool] = []

    flow_type: str = "sequence"

    @model_validator(mode="after")
    def model_valid(self) -> Any:
        if self.manager is None:
            self.manager = Agent(
                name="Team Manager",
                role="Manage the team to complete the task in the best way possible.",
                instructions= [
                    "You are the leader of a team of AI Agents.",
                    "Even though you don't perform tasks by yourself, you have a lot of experience in the field, which allows you to properly evaluate the work of your team members.",
                    "You must always validate the output of the other Agents and you can re-assign the task if you are not satisfied with the result.",
                ],
                tools=self.tools,
                llm=self.llm,
            )
        return self
        
    @classmethod
    def from_yaml(cls, file: str, tools: List[Tool], llm: Optional[Any] = None) -> "Flow":
        if cls == Flow:
            with open(file) as stream:
                try:
                    agents = []
                    content_yaml = yaml.safe_load(stream)
                    for name in content_yaml.get("agents").keys():

                        _values = content_yaml["agents"][name]

                        _tools = []
                        if _values.get("tools"):
                            _agent_tools = _values.get("tools").split(",")
                            for t in tools:
                                if t.name in _agent_tools:
                                    _tools.append(t)

                        context_vars = []
                        if _values.get("context"):
                            context_vars = _values.get("context")                            

                        agent = Agent(
                            name=name,
                            role=_values.get("role"),
                            description=_values.get("description"),
                            response_model=_values.get("response_model"),
                            tools=_tools,
                            context_vars=context_vars,
                            llm=llm,
                        )
                        agents.append(agent)
                    return Flow(agents=agents)
                except Exception as e:
                    logger.error(repr(e))
        raise NotImplementedError(f"from_yaml not implemented for {cls.__name__}")
    
    def describe(self):
        for agent in self.agents:
            print("---")
            print(f"Agent  : {agent.name}")
            if agent.role:
                print(f"Role   : {agent.role}")
            if agent.context_vars:
                print(f"Context:", ",".join(agent.context_vars))
            if agent.tools:
                available_tools = [ tool.name for tool in agent.tools ]
                print(f"Tools  :", ",".join(available_tools))


    def run(self, user_prompt: str, output_file: Optional[str] = None) -> str:

        outputs = {}

        while len(outputs) < len(self.agents):

            for agent in self.agents:

                # check if already run
                if agent.name in outputs.keys():
                    continue

                # check agent dependancies
                if any(outputs.get(var) is None for var in agent.context_vars):
                    continue

                # prepare context and run
                context = None
                if agent.context_vars:
                    context = { f"{var}": outputs[var].content for var in agent.context_vars }
                
                task = None
                if context is None:
                    task = user_prompt # always keep query if no context (first agents)

                outputs[agent.name] = agent.run(task=task, context=context, output_file=output_file)

                logger.debug(f"# {agent.name}\n{ outputs[agent.name].content }", extra={"markup": True})                

        return outputs
    



import json
from typing import List
from pathlib import Path

from gwenflow.types import Document
from gwenflow.readers.base import Reader
from gwenflow.utils import logger


class JSONReader(Reader):

    def read(self, file: Path) -> List[Document]:

        try:

            filename = self.get_file_name(file)
            content  = self.get_file_content(file, text_mode=True)

            json_data = json.loads(content)

            if isinstance(json_data, dict):
                json_data = [json_data]

            documents = []
            for page_num, page_data in enumerate(json_data, start=1):
                content = page_data.pop("content")
                metadata = {"filename": filename, "page": page_num}
                metadata.update(page_data)
                documents.append( Document(
                        id=self.key(f"{filename}_{page_num}"),
                        content=content,
                        metadata=metadata,
                    )
                )
    
        except Exception as e:
            logger.error(f"Error reading file: {e}")
            return []

        return documents




import shutil
import subprocess

def validate_docker_installation() -> None:
    """Check if Docker is installed and running."""
    if not shutil.which("docker"):
        raise RuntimeError(
            f"Docker is not installed. Please install Docker to use code execution with agent: {self.role}"
        )

    try:
        subprocess.run(
            ["docker", "info"],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except subprocess.CalledProcessError:
        raise RuntimeError(
            f"Docker is not running. Please start Docker to use code execution with agent: {self.role}"
        )


from typing import Optional, Union, Any, List, Dict
from pydantic import BaseModel, ConfigDict
from abc import ABC, abstractmethod

import asyncio
import json

from gwenflow.tools import BaseTool
from gwenflow.types import ChatCompletionMessage, ChatCompletionMessageToolCall
from gwenflow.utils import logger


LLM_CONTEXT_WINDOW_SIZES = {
    # openai
    "gpt-4": 8192,
    "gpt-4o": 128000,
    "gpt-4o-mini": 128000,
    "gpt-4-turbo": 128000,
    "o1-preview": 128000,
    "o1-mini": 128000,
    # deepseek
    "deepseek-chat": 128000,
    "deepseek-r1": 128000,
    # google
    "gemma2-9b-it": 8192,
    "gemma-7b-it": 8192,
    # meta
    "llama3-groq-70b-8192-tool-use-preview": 8192,
    "llama3-groq-8b-8192-tool-use-preview": 8192,
    "llama-3.1-70b-versatile": 131072,
    "llama-3.1-8b-instant": 131072,
    "llama-3.2-1b-preview": 8192,
    "llama-3.2-3b-preview": 8192,
    "llama-3.2-11b-text-preview": 8192,
    "llama-3.2-90b-text-preview": 8192,
    "llama3-70b-8192": 8192,
    "llama3-8b-8192": 8192,
    # mistral
    "mixtral-8x7b-32768": 32768,
}

class ChatBase(BaseModel, ABC):
 
    model: str

    system_prompt: Optional[str] = None
    tools: List[BaseTool] = []
    tool_choice: Optional[Union[str, Dict[str, Any]]] = None
    
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def invoke(self, *args, **kwargs) -> Any:
        pass

    @abstractmethod
    async def ainvoke(self, *args, **kwargs) -> Any:
        pass

    @abstractmethod
    def stream(self, *args, **kwargs) -> Any:
        pass
 
    @abstractmethod
    async def astream(self, *args, **kwargs) -> Any:
        pass

    def get_context_window_size(self) -> int:
        # Only using 75% of the context window size to avoid cutting the message in the middle
        return int(LLM_CONTEXT_WINDOW_SIZES.get(self.model, 8192) * 0.75)

    def get_tool_names(self):
        return [tool.name for tool in self.tools]

    def get_tools_map(self):
        return {tool.name: tool for tool in self.tools}


    def handle_tool_call(self, tool_call) -> Dict[str, Any]:

        tool_map  = self.get_tools_map()
        tool_name = tool_call.function.name
                    
        if tool_name not in tool_map.keys():
            logger.error(f"Tool call requested unknown tool '{tool_name}'")
            return None

        try:
            function_args = json.loads(tool_call.function.arguments)
        except json.JSONDecodeError as e:
            logger.error(f"Failed to parse tool arguments: {e}")
            return None

        try:
            logger.debug(f"Tool call: {tool_name}({function_args})")
            observation = tool_map[tool_name].run(**function_args)
            if observation:
                return {
                    "role": "tool",
                    "tool_call_id": tool_call.id,
                    "tool_name": tool_name,
                    "content": f"Observation: {observation}",
                }
        except Exception as e:
            logger.error(f"Error executing tool '{tool_name}': {e}")

        return None

    def handle_tool_calls(self, tool_calls: List[ChatCompletionMessageToolCall]) -> List:
        
        tool_map = self.get_tools_map()
        if not tool_calls or not tool_map:
            return []
        
        messages = []
        for tool_call in tool_calls:
            observation = self.handle_tool_call(tool_call)
            if observation:
                messages.append(observation)
            
        return messages

    async def ahandle_tool_calls(self, tool_calls: List[ChatCompletionMessageToolCall]) -> List:
        
        tool_map = self.get_tools_map()
        if not tool_calls or not tool_map:
            return []

        tasks = []
        for tool_call in tool_calls:
            # task = asyncio.create_task(self.handle_tool_call(tool_call))
            task = asyncio.create_task(asyncio.to_thread(self.handle_tool_call, tool_call))
            tasks.append(task)

        messages = []

        results = await asyncio.gather(*tasks)
        for observation in results:
            if observation:
                messages.append(observation)

        return messages
    




import logging
import hashlib
from typing import Optional
from enum import Enum

from qdrant_client import QdrantClient
from qdrant_client.models import (
    Distance,
    FieldCondition,
    Filter,
    MatchValue,
    PointIdsList,
    PointStruct,
    Range,
    DatetimeRange,
    VectorParams,
    CreateAliasOperation,
    CreateAlias,
    DeleteAliasOperation,
    DeleteAlias,
    PayloadSchemaType)

from gwenflow.vector_stores.base import VectorStoreBase
from gwenflow.embeddings import Embeddings, GwenlakeEmbeddings
from gwenflow.reranker import Reranker
from gwenflow.types import Document

logger = logging.getLogger(__name__)

class Index(str, Enum):
    keyword = PayloadSchemaType.KEYWORD
    integer = PayloadSchemaType.INTEGER
    float = PayloadSchemaType.FLOAT
    bool = PayloadSchemaType.BOOL
    geo = PayloadSchemaType.GEO
    datetime = PayloadSchemaType.DATETIME
    text = PayloadSchemaType.TEXT

class Qdrant(VectorStoreBase):

    def __init__(
        self,
        collection: str,
        embeddings: Embeddings = GwenlakeEmbeddings(),
        distance: Distance = Distance.COSINE,
        client: QdrantClient = None,
        host: str = None,
        port: int = 6333,
        path: str = None,
        url: str = None,
        api_key: str = None,
        reranker: Optional[Reranker] = None,
        on_disk: bool = True,
    ):
        """
        Initialize the Qdrant vector store.

        Args:
            collection (str): Name of the collection.
            client (QdrantClient, optional): Existing Qdrant client instance. Defaults to None.
            host (str, optional): Host address for Qdrant server. Defaults to None.
            port (int, optional): Port for Qdrant server. Defaults to None.
            path (str, optional): Path for local Qdrant database. Defaults to None.
            url (str, optional): Full URL for Qdrant server. Defaults to None.
            api_key (str, optional): API key for Qdrant server. Defaults to None.
        """

        # Embedder
        self.embeddings = embeddings

        # Distance metric
        self.distance = distance

        # reranker
        self.reranker = reranker

        # on disk:
        self.on_disk = on_disk

        if client:
            self.client = client
        else:
            params = {}
            if api_key:
                params["api_key"] = api_key
            if url:
                params["url"] = url
            if host and port:
                params["host"] = host
                params["port"] = port
            if not params:
                params["path"] = path
            # if self.on_disk:
            #     params = { "url": "http://localhost:6333" }

            self.client = QdrantClient(**params)

        self.collection = collection
        self.create()

    def get_collections(self) -> list:
        """
        List all collections.

        Returns:
            list: List of collection names.
        """
        try:
            return self.client.get_collections()
        except Exception as e:
            logger.error(f"Error while reading collections: {e}")
        return []

    def get_aliases(self) -> list:
            """
            List all aliases.

            Returns:
                list: List of alias names.
            """
            try:
                return self.client.get_aliases()
            except Exception as e:
                logger.error(f"Error while reading aliases: {e}")
            return []


    def create(self):
        """Create collection."""
        # Skip creating collection if already exists or an existing alias is already present
        collections = self.get_collections()
        if collections:
            for collection in collections.collections:
                if collection.name == self.collection:
                    logging.debug(f"Collection {self.collection} already exists. Skipping creation.")
                    return

        aliases = self.get_aliases()
        if aliases:
            for alias in aliases.aliases:
                if alias.alias_name == self.collection:
                    logging.debug(f"Alias {self.collection} already exists. Skipping creation.")
                    return
        try:
            self.client.create_collection(
                collection_name=self.collection,
                vectors_config=VectorParams(size=self.embeddings.dimensions, distance=self.distance, on_disk=self.on_disk),
            )
        except Exception as e:
            logger.error(f"Error while creating collection: {e}")

    def drop(self):
        """Drop collection."""
        self.client.delete_collection(collection_name=self.collection)

    def count(self) -> int:
        result = self.client.count(collection_name=self.collection, exact=True)
        return result.count

    def info(self) -> dict:
        """
        Get information about the collection.

        Returns:
            dict: Collection information.
        """
        return self.client.get_collection(collection_name=self.collection)

    def insert(self, documents: list[Document]):
        """
        Insert documents into a collection.

        Args:
            documents (list): List of documents to insert.
        """
        logger.info(f"Inserting {len(documents)} documents into collection {self.collection}")
        _embeddings = self.embeddings.embed_documents([document.content for document in documents])
        points = []
        for document, embedding in zip(documents, _embeddings):
            if document.id is None:
                document.id = hashlib.md5(document.content.encode(), usedforsecurity=False).hexdigest()
            _id = document.id
            _payload = document.metadata
            _payload["content"] = document.content
            points.append(
                PointStruct(
                    id=_id,
                    vector=embedding,
                    payload=_payload,
                )
            )
    
        if len(points) > 0:
            self.client.upsert(collection_name=self.collection, points=points)

    def _create_filter(self, filters: dict) -> Filter:
        """
        Create a Filter object from the provided filters.

        Args:
            filters (dict): Filters to apply.

        Returns:
            Filter: The created Filter object.
        """
        conditions = []
        collection_payload = self.client.get_collection(collection_name=self.collection).payload_schema
        for key, value in filters.items():
            if isinstance(value, dict) and "gte" in value and "lte" in value:
                if key in collection_payload and collection_payload[key].data_type=="datetime":
                    conditions.append(FieldCondition(key=key, range=DatetimeRange(gte=value["gte"], lte=value["lte"])))
                else:
                    conditions.append(FieldCondition(key=key, range=Range(gte=value["gte"], lte=value["lte"])))
            else:
                conditions.append(FieldCondition(key=key, match=MatchValue(value=value)))
        return Filter(must=conditions) if conditions else None

    def search(self, query: str, limit: int = 5, filters: dict = None) -> list[Document]:
        """
        Search for similar vectors.

        Args:
            query (str): Query.
            limit (int, optional): Number of results to return. Defaults to 5.
            filters (dict, optional): Filters to apply to the search. Defaults to None.

        Returns:
            list: Search results.
        """

        query_embedding = self.embeddings.embed_query(query)
        if query_embedding is None:
            logger.error(f"Error getting embedding for Query: {query}")
            return []

        query_filter = self._create_filter(filters) if filters else None
        hits = self.client.search(
            collection_name=self.collection,
            query_vector=query_embedding,
            query_filter=query_filter,
            limit=limit,
            with_payload=True,
            with_vectors=False,
        )
 
        documents = []
        for d in hits:

            if d.payload is None:
                continue

            content = None
            if "content" in d.payload:
                content = d.payload.pop("content")
            elif "chunk" in d.payload:
                content = d.payload.pop("chunk")

            # Note: should be removed
            if "metadata" in d.payload:
                d.payload = d.payload["metadata"]

            if isinstance(d.id, int):
                d.id = str(d.id)
                

            documents.append(
                Document(
                    id=d.id,
                    content=content,
                    metadata=d.payload,
                    score=1-d.score,
                )
            )
    
        if self.reranker:
            documents = self.reranker.rerank(query=query, documents=documents)

        return documents

    def delete(self, id: int):
        """
        Delete a vector by ID.

        Args:
            id (int): ID of the vector to delete.
        """
        self.client.delete(
            collection_name=self.collection,
            points_selector=PointIdsList(
                points=[id],
            ),
        )

    def get(self, id: int) -> dict:
        """
        Retrieve a vector by ID.

        Args:
            id (int): ID of the vector to retrieve.

        Returns:
            dict: Retrieved vector.
        """
        result = self.client.retrieve(collection_name=self.collection, ids=[id], with_payload=True)
        return result[0] if result else None


    def list(self, filters: dict = None, limit: int = 100) -> list:
        """
        List all vectors in a collection.

        Args:
            filters (dict, optional): Filters to apply to the list. Defaults to None.
            limit (int, optional): Number of vectors to return. Defaults to 100.

        Returns:
            list: List of vectors.
        """
        query_filter = self._create_filter(filters) if filters else None
        result = self.client.scroll(
            collection_name=self.collection,
            scroll_filter=query_filter,
            limit=limit,
            with_payload=True,
            with_vectors=False,
        )
        return result
    
    def create_alias(self, alias_name: str):
        """
        Create an alias for a list of collections.

        Args:
            alias (str): Name of the alias.
            collections (List[str]): List of collection names.
        """
        self.client.update_collection_aliases(
        change_aliases_operations=[
            CreateAliasOperation(
                create_alias=CreateAlias(
                    collection_name=self.collection, alias_name=alias_name
                )
            )
        ]
    )
        
    def delete_alias(self, alias_name: str):
        """
        Delete an alias.

        Args:
            alias (str): Name of the alias.
        """
        self.client.update_collection_aliases(
        change_aliases_operations=[
            DeleteAliasOperation(
                delete_alias=DeleteAlias(alias_name=alias_name)
            ),
        ]
    )
        
    def switch_alias(self, alias_name: str):
        """
        Switch an alias to a new collection.

        Args:
            alias (str): Name of the alias.
            collection (str): Name of the collection.
        """
        self.client.update_collection_aliases(
        change_aliases_operations=[
            DeleteAliasOperation(
                delete_alias=DeleteAlias(alias_name=alias_name)
            ),
            CreateAliasOperation(
                create_alias=CreateAlias(
                    collection_name=self.collection, alias_name=alias_name
                )
            )
        ]
    )
        
    def add_index(self, field_name: str, index_type: str):
        """
        Args:
            field_name (str): Name of the field to index.
            index_type (str): Type of index (must be one of the valid types).

        Raises:
            ValueError: If the index type is not valid.
        """

        if index_type not in Index.__members__:
            raise ValueError(f"Invalid index_type: {index_type}. Must be one of {list(Index.__members__.keys())}")
        
        self.client.create_payload_index(
            collection_name=self.collection,
            field_name=field_name,
            field_schema=Index(index_type)
        )
         


from gwenflow.tools.base import BaseTool, Tool
from gwenflow.tools.duckduckgo import DuckDuckGoSearchTool, DuckDuckGoNewsTool
from gwenflow.tools.pdf import PDFTool
from gwenflow.tools.website import WebsiteTool
from gwenflow.tools.wikipedia import WikipediaTool
from gwenflow.tools.yahoofinance import (
    YahooFinanceNews,
    YahooFinanceStock,
    YahooFinanceScreen,
)
from gwenflow.tools.tavily import TavilyWebSearchTool

__all__ = [
    "BaseTool",
    "Tool",
    "WikipediaTool",
    "WebsiteTool",
    "PDFTool",
    "DuckDuckGoSearchTool",
    "DuckDuckGoNewsTool",
    "YahooFinanceNews",
    "YahooFinanceStock",
    "YahooFinanceScreen",
    "TavilyWebSearchTool",
]



from typing import Optional, List

from pydantic import BaseModel, ConfigDict


class Embeddings(BaseModel):
    """Base class for embeddings"""

    model: str
    dimensions: Optional[int] = 1536

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        raise NotImplementedError

    def embed_query(self, text: str) -> List[float]:
        raise NotImplementedError




from typing import List, Callable, Union, Optional, Any, Dict
from pydantic import BaseModel
import json

from gwenflow.agents import Agent
from gwenflow.flows import Flow
from gwenflow.tools import BaseTool
from gwenflow.utils.json import parse_json_markdown
from gwenflow.utils import logger


EXAMPLE = [
    {
        "name": "Biographer",
        "role": "Write two paragraphs of max 500 words each",
        "tools": ["wikipedia"],
        "context": [],
    },
    { 
        "name": "Summarizer",
        "role": "List of 10 bullet points",
        "tools": None,
        "context": ["Biographer"],
    },
    {
        "name": "RelatedTopics",
        "role": "Generate a list of 5 related topics",
        "tools": None,
        "context": ["Summarizer"],
    },
    {
        "name": "Final Report",
        "role": "Produce a final report in a Powerpoint file (pptx format)",
        "tools": ["wikipedia","python"],
        "context": ["Biographer","Summarizer","RelatedTopics"],
    }
]


TASK_GENERATOR = """
You are an AI Agent creator tasked with generating a list of AI Agents based on a given set of tasks. \
    Your goal is to create a JSON array of Agent objects, each designed to accomplish a specific task within a larger objective.

Available Tools:
<tools>
{tools}
</tools>

Tasks to Accomplish:
<tasks>
{tasks}
</tasks>

Instructions for Creating AI Agents:

1. Analyze the given tasks and break them down into discrete steps that can be assigned to individual AI Agents.

2. For each task, create an Agent with the following properties:
   - name: A descriptive name for the Agent
   - role: A descriptive role for the Agent
   - task: A detailed description of what the Agent needs to accomplish
   - context: An array of Agent names whose results this Agent depends on (can be empty)
   - tools: An array of tools the Agent can use to complete its task (can be empty)
   - description: The task formulated as a question to ask the user

3. Adhere to these guidelines:
   - Create Agents based on the tasks provided
   - Limit Agents to those that can be completed with the available tools listed above
   - Ensure all tasks are detailed and specific
   - When multiple searches are required, use the tools multiple times
   - Do not use the same tool with the same arguments multiple times
   - Make sure all tasks are in chronological order
   - Validate the output of other Agents and reassign tasks if necessary

4. Wrap the following process in <agent_creation_process> tags for each Agent:
   a. List out all tasks and number them for organization
   b. Identify the specific task the Agent needs to accomplish
   c. Determine a suitable name for the Agent
   d. Write a detailed task description
   e. Consider and list potential dependencies on other tasks/Agents
   f. Identify any dependencies on other Agents (context)
   g. Select appropriate tools from the available options
   h. Formulate the task description as a question for the user

5. After creating all Agents, review the entire list to ensure:
   - All tasks are covered
   - Agents are in the correct chronological order
   - Dependencies (context) are correctly specified
   - Tool usage is appropriate and necessary

6. Generate the final JSON array of Agent objects.

Output Format:
The output should be a JSON array of Agent objects. Here's a generic example of the structure:

```json
{examples}
```

Remember to replace the generic content with actual Agent data based on the given tasks.

Now, please create the list of AI Agents based on the provided tasks.
"""

class AutoFlow(Flow):

    def run(self, user_prompt: str, output_file: Optional[str] = None) -> str:

        tools = [ tool.name for tool in self.tools ]
        tools = ", ".join(tools)

        task_prompt = TASK_GENERATOR.format(tasks=user_prompt, tools=tools, examples=json.dumps(EXAMPLE, indent=4))
        agents_json = self.llm.invoke(messages=[{"role": "user", "content": task_prompt}])
        agents_json = parse_json_markdown(agents_json)
        
        for agent_json in agents_json:

            tools = []
            if agent_json.get("tools"):
                for t in self.tools:
                    if t.name in agent_json["tools"]:
                        tools.append(t)

            agent = Agent(
                llm=self.llm,
                name=agent_json.get("name"),
                role=agent_json.get("role"),
                tools=tools,
                context_vars=agent_json.get("context"),
            )

            self.agents.append(agent)

        self.describe()

        return super().run(user_prompt=user_prompt, output_file=output_file)



import json
from typing import Any, Optional
from pydantic import Field, model_validator

from gwenflow.tools import BaseTool


class DuckDuckGoBaseTool(BaseTool):

    region: Optional[str] = "wt-wt"
    source: str = "text"
    time: Optional[str] = "y"
    max_results: int = 5
    safesearch: str = "moderate"
    backend: str = "api"

    @model_validator(mode="before")
    @classmethod
    def validate_environment(cls, values: Any) -> Any:
        """Validate that the python package exists in environment."""
        try:
            from duckduckgo_search import DDGS  # noqa: F401
        except ImportError:
            raise ImportError("duckduckgo-search is not installed. Please install it with `pip install duckduckgo-search`.")
        return values


class DuckDuckGoSearchTool(DuckDuckGoBaseTool):

    name: str = "DuckDuckGoSearchTool"
    description: str = "Search for a query in DuckDuckGo and returns the content."

    def _run(self, query: str = Field(description="The search query.")):
        from duckduckgo_search import DDGS
        with DDGS() as ddgs:
            results = ddgs.text(
                query,
                region=self.region,  # type: ignore[arg-type]
                safesearch=self.safesearch,
                timelimit=self.time,
                max_results=self.max_results,
                backend=self.backend,
            )
            return json.dumps(results)


class DuckDuckGoNewsTool(DuckDuckGoBaseTool):

    name: str = "DuckDuckGoNewsTool"
    description: str = "Search for a query in DuckDuckGo News and returns the content."

    def _run(self, query: str = Field(description="The search query.")) -> str:
        from duckduckgo_search import DDGS
        with DDGS() as ddgs:
            results = ddgs.news(
                query,
                region=self.region,  # type: ignore[arg-type]
                safesearch=self.safesearch,
                timelimit=self.time,
                max_results=self.max_results,
            )
            return json.dumps(results)



from pydantic import BaseModel, ConfigDict, Field, SecretStr, model_validator
from typing import Any, Dict, List, Optional, cast
import os
import requests

from gwenflow.types import Document
from gwenflow.reranker.base import Reranker


class GwenlakeReranker(Reranker):
    """Gwenlake reranker."""

    api_base: str = "https://api.gwenlake.com/v1/rerank"
    api_key: Optional[SecretStr] = None
    
    model_config = ConfigDict(
        extra="forbid",
    )


    @model_validator(mode="before")
    @classmethod
    def validate_environment(cls, values: Dict) -> Dict:
        values["api_key"] = os.getenv("GWENLAKE_API_KEY")
        if "model" not in values:
            values["model"] = "BAAI/bge-reranker-v2-m3"
        return values

    def _rerank(self, query: str, input: List[str]) -> List[List[float]]:

        api_key = cast(SecretStr, self.api_key).get_secret_value()

        payload = {"query": query, "input": input, "model": self.model}

        # HTTP headers for authorization
        headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
        }

        # send request
        try:
            response = requests.post(self.api_base, headers=headers, json=payload)
        except requests.exceptions.RequestException as e:
            raise ValueError(f"Error raised by inference endpoint: {e}")
        
        if response.status_code != 200:
            raise ValueError(
                f"Error raised by inference API: rate limit exceeded.\nResponse: "
                f"{response.text}"
            )

        parsed_response = response.json()
        if "data" not in parsed_response:
            raise ValueError("Error raised by inference API.")

        reranking = []
        for e in parsed_response["data"]:
            reranking.append(e)
        
        return reranking

    def rerank(self, query: str, documents: List[Document]) -> List[Document]:

        if not documents:
            return []
        
        batch_size = 100
        reranked_documents = []
        try:
            for i in range(0, len(documents), batch_size):
                i_end = min(len(documents), i+batch_size)
                batch = documents[i:i_end]
                batch_processed = []
                for document in batch:
                    batch_processed.append(document.content)
                reranked_documents += self._rerank(query=query, input=batch_processed)
        except Exception as e:
            print(repr(e))
            return None

        if len(reranked_documents) > 0:

            compressed_documents = documents.copy()

            for i, _ in enumerate(compressed_documents):
                compressed_documents[i].score = reranked_documents[i]["relevance_score"]

            # Order by relevance score
            compressed_documents.sort(
                key=lambda x: x.score if x.score is not None else float("-inf"),
                reverse=True,
            )

            if self.top_k is not None:
                compressed_documents = compressed_documents[:self.top_k]

            if self.threshold is not None:
                compressed_documents = [d for d in compressed_documents if d.score > self.threshold]

            return compressed_documents

        return []



from typing import Any, Dict
import os
import logging

from gwenflow.llms.openai import ChatOpenAI


logger = logging.getLogger(__name__)


class ChatMistralAI(ChatOpenAI):
 
    model: str = "open-mistral-7b"
    base_url: str = "https://api.mistral.ai"

    def _get_client_params(self) -> Dict[str, Any]:

        api_key = self.api_key
        if api_key is None:
            api_key = os.environ.get("MISTRAL_API_KEY")
        if api_key is None:
            raise ValueError(
                "The api_key client option must be set either by passing api_key to the client or by setting the MISTRAL_API_KEY environment variable"
            )

        client_params = {
            "api_key": api_key,
            "base_url": self.base_url,
            "timeout": self.timeout,
            "max_retries": self.max_retries,
        }

        client_params = {k: v for k, v in client_params.items() if v is not None}

        return client_params
     



def merge_fields(target, source):
    for key, value in source.items():
        if isinstance(value, str):
            target[key] += value
        elif value is not None and isinstance(value, dict):
            merge_fields(target[key], value)


def merge_chunk(final_response: dict, delta: dict) -> None:
    delta.pop("role", None)
    merge_fields(final_response, delta)

    tool_calls = delta.get("tool_calls")
    if tool_calls and len(tool_calls) > 0:
        index = tool_calls[0].pop("index")
        merge_fields(final_response["tool_calls"][index], tool_calls[0])




from pydantic import BaseModel, Field
from typing import Any, Dict, Optional, Union
from collections.abc import Sequence


class ChatMessage(BaseModel):
    """Chat message class."""

    content: Union[str, list[Union[str, dict]]]
    role: str
    name: Optional[str] = None
    meta: Dict[str, Any] = Field(default_factory=dict, hash=False)

    def __str__(self) -> str:
        return f"{self.role}: {self.content}"

    def to_dict(self, **kwargs: Any) -> Dict[str, Any]:
        return self.model_dump(**kwargs)
    
    def to_openai(self) -> Dict[str, Any]:
        return { "role": self.role, "content": self.content }


def messages_to_dict(messages: Sequence[ChatMessage]) -> list[dict]:
    """Convert a sequence of Messages to a list of dictionaries.

    Args:
        messages: Sequence of messages (as ChatMessages) to convert.

    Returns:
        List of messages as dicts.
    """
    return [m.to_dict() for m in messages]


def messages_to_openai(messages: Sequence[ChatMessage]) -> list[dict]:
    """Convert a sequence of Messages to a list of dictionaries in OpenAI format.

    Args:
        messages: Sequence of messages (as ChatMessages) to convert.

    Returns:
        List of messages as dicts.
    """
    return [m.to_openai() for m in messages]



import os
from typing import Any, Optional
from pydantic import Field, model_validator

from gwenflow.tools import BaseTool
from gwenflow.utils import logger


class TavilyBaseTool(BaseTool):

    client: Optional[Any] = None
    api_key: Optional[str] = None

    @model_validator(mode="after")
    def validate_environment(self) -> 'TavilyBaseTool':
        """Validate that the python package exists in environment."""
        try:
            from tavily import TavilyClient
            if self.client is None:
                if self.api_key is None:
                    self.api_key = os.getenv("TAVILY_API_KEY")
                if self.api_key is None:
                    logger.error("TAVILY_API_KEY not provided")
                self.client = TavilyClient(api_key=self.api_key)
        except ImportError:
            raise ImportError("`tavily-python` not installed. Please install using `pip install tavily-python`")
        return self

class TavilyWebSearchTool(TavilyBaseTool):

    name: str = "TavilyWebSearchTool"
    description: str = "Use this function to search the web for a given query."

    def _run(self, query: str = Field(description="Query to search for.")) -> str:
        search_depth = "advanced"
        max_tokens = 6000
        return self.client.get_search_context(query=query, search_depth=search_depth, max_tokens=max_tokens)



import re
import json

def parse_json_markdown(json_string: str) -> dict:
    # Remove the triple backticks if present
    json_string = json_string.strip()
    start_index = json_string.find("```json")
    end_index = json_string.find("```", start_index + len("```json"))

    if start_index != -1 and end_index != -1:
        extracted_content = json_string[start_index + len("```json"):end_index].strip()
        
        # Parse the JSON string into a Python dictionary
        parsed = json.loads(extracted_content)
    elif start_index != -1 and end_index == -1 and json_string.endswith("``"):
        end_index = json_string.find("``", start_index + len("```json"))
        extracted_content = json_string[start_index + len("```json"):end_index].strip()
        
        # Parse the JSON string into a Python dictionary
        parsed = json.loads(extracted_content)
    elif json_string.startswith("{"):
        # Parse the JSON string into a Python dictionary
        parsed = json.loads(json_string)
    else:
        raise Exception("Could not find JSON block in the output.")

    return parsed

def parse_and_check_json_markdown(text: str, expected_keys: list[str]) -> dict:
    try:
        json_obj = parse_json_markdown(text)
    except json.JSONDecodeError as e:
        raise Exception(f"Got invalid JSON object. Error: {e}")
    for key in expected_keys:
        if key not in json_obj:
            raise Exception(
                f"Got invalid return object. Expected key `{key}` "
                f"to be present, but got {json_obj}"
            )
    return json_obj

from datetime import date, datetime

def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError ("Type %s not serializable" % type(obj))

def to_json(obj):
    return json.dumps(obj, default=json_serial)

def extract_json_str(text: str) -> str:
    """Extract JSON string from text."""
    # NOTE: this regex parsing is taken from langchain.output_parsers.pydantic
    match = re.search(r"\{.*\}", text.strip(), re.MULTILINE | re.IGNORECASE | re.DOTALL)
    if not match:
        raise ValueError(f"Could not extract json string from output: {text}")

    return match.group()




from typing import List
from pathlib import Path

from gwenflow.types import Document
from gwenflow.readers.base import Reader
from gwenflow.utils import logger



class TextReader(Reader):

    def read(self, file: Path) -> List[Document]:

        try:

            filename = self.get_file_name(file)
            content  = self.get_file_content(file, text_mode=True)

            documents = [
                Document(
                    id=self.key(filename),
                    content=content,
                    metadata={"filename": filename},
                )
            ]

        except Exception as e:
            logger.error(f"Error reading file: {e}")
            return []
        
        return documents


import json
import re
from typing import Union, Optional, Any, Dict, List, Iterator
from collections import defaultdict

from gwenflow.types import ChatCompletionMessage
from gwenflow.agents.types import AgentResponse
from gwenflow.agents.agent import Agent
from gwenflow.agents.react.types import ActionReasoningStep
from gwenflow.agents.react.parser import ReActOutputParser
from gwenflow.agents.react.prompts import PROMPT_REACT
from gwenflow.agents.prompts import PROMPT_TASK, PROMPT_TOOLS
from gwenflow.agents.utils import merge_chunk
from gwenflow.utils import logger


MAX_TURNS = float('inf')


class ReActAgent(Agent):

    is_react: bool = True
    description: str = "You are a meticulous and thoughtful assistant that solves a problem by thinking through it step-by-step."
    reasoning_steps: List[ActionReasoningStep] = []

    def parse(self, text: str) -> ActionReasoningStep:
        return ReActOutputParser().parse(text)

    def get_system_message(self, context: Optional[Any] = None):
        """Return the system message for the Agent."""

        # Add additional instructions
        additional_guidelines = [
            "Your goal is to reason about the task or query and decide on the best course of action to answer it accurately.",
            "If you cannot find the necessary information after using available tools, admit that you don't have enough information to answer the query confidently.",
        ]
        self.instructions = additional_guidelines + self.instructions

        # default system message
        system_message = super(ReActAgent, self).get_system_message(context=context)

        # ReAct
        tool_names = ",".join(self.get_tool_names())
        prompt = PROMPT_REACT.format(tool_names=tool_names).strip()
        system_message["content"] += f"\n\n{prompt}"

        return system_message

    def handle_tool_call(self, reasoning_step: ActionReasoningStep) -> Dict:
        
        tool_map = self.get_tools_map()

        # handle missing tool case, skip to next tool
        if reasoning_step.action not in tool_map:
            # logger.warning(f"Unknown tool {reasoning_step.action}, should be instead one of { tool_map.keys() }.")
            return {
                "role": "user",
                "content": "Ok. Let’s proceed with the next step.",
            }

        observation = self.execute_tool_call(reasoning_step.action, reasoning_step.action_input)
                
        return {
            "role": "user",
            "content": f"Observation: {observation}",
        }
    
    def invoke(self, messages: list, stream: bool = False) ->  Union[Any, Iterator[Any]]:

        params = {
            "messages": messages,
            "parse_response": False,
        }

        response_format = None
        if self.response_model:
            response_format = {"type": "json_object"}

        if stream:
            return self.llm.stream(**params, response_format=response_format)
        
        return self.llm.invoke(**params, response_format=response_format)

    def reason(self, task: str):

        if self.reasoning_model is None:
            return None
        
        user_prompt = ""
        
        if self.tools:
            tools = self.get_tools_text_schema()
            user_prompt += PROMPT_TOOLS.format(tools=tools).strip() + "\n\n"

        user_prompt += PROMPT_TASK.format(task=task).strip()
        user_prompt += "\n\nPlease help me with some thoughts, steps and guidelines to answer accurately and precisely to this task."

        params = {
            "messages": [{"role": "user", "content": user_prompt}],
            "parse_response": True,
        }

        logger.debug("Reasoning.")
        response = self.reasoning_model.invoke(**params)

        # only keep text outside <think>
        reasoning_content = re.sub(r'<think>.*?</think>', '', response, flags=re.DOTALL)
        if not reasoning_content:
            return None
        
        reasoning_content = reasoning_content.strip()
        
        logger.debug("Thought: " + reasoning_content)

        return reasoning_content

    def _run(
        self,
        task: Optional[str] = None,
        *,
        context: Optional[Any] = None,
        stream: Optional[bool] = False,
    ) ->  Iterator[AgentResponse]:

        # add reasoning
        if self.reasoning_model:
            reasoning_message = self.reason(task=task)
            if reasoning_message:
                if not context:
                    context = {}
                if isinstance(context, str):
                    text = context
                    context = { "context": text }
                context["thinking"] = reasoning_message

        # messages for model
        messages_for_model = []
        system_message = self.get_system_message(context=context)
        user_message = self.get_user_message(task=task, context=context)

        # check if system prompt is allow and add messages to messages_for_model
        if self.system_prompt_allowed:
            if system_message:
                messages_for_model.append(system_message)
            if user_message:
                messages_for_model.append(user_message)
                self.memory.add_message(user_message)
        else:
            system_message["role"] = "user"
            if user_message:
                system_message["content"] += "\n\n" + user_message["content"]
            messages_for_model.append(system_message)
            self.memory.add_message(system_message)
        # global loop
        init_len = len(messages_for_model)
        while len(messages_for_model) - init_len < MAX_TURNS:

            if stream:
                message = {
                    "content": "",
                    "sender": self.name,
                    "role": "assistant",
                }

                completion = self.invoke(messages=messages_for_model, stream=True)

                for chunk in completion:
                    if len(chunk.choices) > 0:
                        delta = json.loads(chunk.choices[0].delta.json())
                        if delta["role"] == "assistant":
                            delta["sender"] = self.name
                        if delta["content"]:
                            yield AgentResponse(
                                delta=delta["content"],
                                messages=None,
                                agent=self,
                                tools=self.tools,
                            )
                        delta.pop("role", None)
                        delta.pop("sender", None)
                        merge_chunk(message, delta)

                message = ChatCompletionMessage(**message)
            
            else:
                completion = self.invoke(messages=messages_for_model)                
                message = completion.choices[0].message
                message.sender = self.name

            # add messages to the current message stack
            message_dict = json.loads(message.model_dump_json())
            messages_for_model.append(message_dict)

            # parse response
            reasoning_step = self.parse(message_dict["content"])
            self.reasoning_steps.append(reasoning_step)

            # show response
            logger.debug(f"\n---\nThought: { reasoning_step.thought }")

            # final answer ?
            if reasoning_step.is_done:
                logger.debug("Task done.")
                self.memory.add_message(message_dict)
                break

            # handle tool calls
            observation = self.handle_tool_call(reasoning_step)
            if observation:
                messages_for_model.append(observation)

        content = messages_for_model[-1]["content"]
        if self.response_model:
            content = json.loads(content)

        if len(self.reasoning_steps)>0:
            last_reasoning_step = self.reasoning_steps[-1]
            content = last_reasoning_step.thought + "\n\n" + last_reasoning_step.response

        yield AgentResponse(
            content=content,
            messages=messages_for_model[init_len:],
            agent=self,
            tools=self.tools,
        )

    def get_reasoning_steps(self) -> str:
        steps = []
        steps.append("# Reasoning Steps")
        for i, reasoning_step in enumerate(self.reasoning_steps):
            if reasoning_step.is_done:
                text  = f"## Final Step\n"
                text += f"- **Reasoning:** {reasoning_step.thought}\n"
                text += f"- **Final Answer:** {reasoning_step.response}\n"
                steps.append(text)
                break
            else:
                text  = f"## Step {i+1}.\n"
                text += f"- **Reasoning:** {reasoning_step.thought}\n"
                text += f"- **Action:** {reasoning_step.action}, **Inputs:** {reasoning_step.action_input}\n"
                steps.append(text)

        steps = "\n\n".join(steps)
        return steps




from __future__ import annotations

from pydantic import BaseModel, model_validator
from typing import Any, Optional, Union

from gwenflow.prompts.template import PromptTemplate


class PipelinePromptTemplate(BaseModel):
    """Pipeline prompt template for a language model."""

    prompts: list[PromptTemplate]
    """The pipeline."""

    input_variables: list[str]
    """The prompt input variables."""
    
    @model_validator(mode="before")
    @classmethod
    def get_input_variables(cls, values: dict) -> Any:
        """Get input variables."""
        all_variables = set()
        for prompt in values["prompts"]:
            all_variables.update(prompt.input_variables)
        values["input_variables"] = list(all_variables)
        return values


    def format(self, **kwargs: Any) -> list:
        """Format the prompt with the inputs.

        Args:
            kwargs: Any arguments to be passed to the prompt template.

        Returns:
            A formatted string.
        """
        _formatted = []
        for prompt in self.prompts:
            _formatted.append( prompt.format(**kwargs) )
        return _formatted


from typing import Any, List, Union
from pydantic import BaseModel, Field, ConfigDict

import io
import requests
import hashlib
from typing import List
from pathlib import Path

from gwenflow.utils import logger
from gwenflow.types import Document
from gwenflow.utils.aws import aws_s3_read_file, aws_s3_read_text_file, aws_s3_uri_to_bucket_key


class Reader(BaseModel):

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def key(self, text) -> str:
        return hashlib.md5(text.encode(), usedforsecurity=False).hexdigest()

    def read(self, obj: Any) -> List[Document]:
        raise NotImplementedError

    def get_file_name(self, file: Union[Path, io.BytesIO]):
        if isinstance(file, io.BytesIO):
            return "noname"
        if not isinstance(file, Path):
            return str(Path(file))
        return str(file)

    def get_file_content(self, file: Union[Path, io.BytesIO], text_mode: bool=False):

        try:

            if isinstance(file, io.BytesIO):
                return file

            filename = str(file)

            if filename.startswith("s3://"):
                bucket, key = aws_s3_uri_to_bucket_key(file)
                if text_mode:
                    return aws_s3_read_text_file(bucket, key)
                return aws_s3_read_file(bucket, key)

            elif filename.startswith("http://") or filename.startswith("https://"):
                response = requests.get(str(file))
                if text_mode:
                    return response.text
                return io.BytesIO(response.content)
            
            else:
                if not isinstance(file, Path):
                    file = Path(file)
                if not file.exists():
                    raise FileNotFoundError(f"Could not find file: {file}")
                if text_mode:
                    return file.read_text("utf-8")
                return io.BytesIO(file.read_bytes())
        
        except Exception as e:
            logger.error(f"Error reading file: {e}")
        
        return None



import uuid
from typing import List, Callable, Union, Optional, Any, Dict, Iterator, Literal, Sequence, overload, Type
from pydantic import BaseModel, model_validator, field_validator, Field

from gwenflow.types import ChatMessage


class BaseChatMemory(BaseModel):
 
    id: Optional[str] = Field(None, validate_default=True)
    messages: list[ChatMessage] = []

    @field_validator("id", mode="before")
    def set_id(cls, v: Optional[str]) -> str:
        id = v or str(uuid.uuid4())
        return id

    def to_string(self) -> str:
        """Convert memory to string."""
        return self.json()

    def to_dict(self, **kwargs: Any) -> dict:
        """Convert memory to dict."""
        return self.dict()
    
    def reset(self):
        self.messages = []

    def get_all(self):
        return self.messages

    def add_messages(self, messages: list[ChatMessage]):
        for message in messages:
            self.messages.append(ChatMessage(**message))

    def add_message(self, message: ChatMessage):
        self.messages.append(ChatMessage(**message))





import time
import random
import httpx
from urllib.parse import urljoin, urlparse

from typing import Set, Dict, List, Tuple

from gwenflow.readers import PDFReader
from gwenflow.readers.base import Reader
from gwenflow.types import Document
from gwenflow.readers.base import Reader
from gwenflow.utils import logger

try:
    from bs4 import BeautifulSoup  # noqa: F401
except ImportError:
    raise ImportError("BeautifulSoup is not installed. Please install it with `pip install beautifulsoup4`.")


class WebsiteReader(Reader):
    """Reader for Websites"""

    max_depth: int = 3
    max_links: int = 100000

    delay: bool = True

    _visited: Set[str] = set()
    _urls_to_crawl: List[Tuple[str, int]] = []

    def sleep(self, min_seconds=1, max_seconds=3):
        """Introduce a random delay."""
        sleep_time = random.uniform(min_seconds, max_seconds)
        time.sleep(sleep_time)

    def _get_primary_domain(self, url: str) -> str:
        """Extract primary domain from the given URL (excluding subdomains)."""
        domain_parts = urlparse(url).netloc.split(".")
        return ".".join(domain_parts[-2:])

    def _extract_html_content(self, soup: BeautifulSoup) -> str:
        """Extracts the main content from a BeautifulSoup object."""

        for tag in ["article", "main"]:
            element = soup.find(tag)
            if element:
                return element.get_text(strip=True, separator=" ")

        for class_name in ["content", "main-content", "page-content", "post-content"]:
            element = soup.find(class_=class_name)
            if element:
                return element.get_text(strip=True, separator=" ")

        return ""

    def _extract_pdf_content(self, url: str) -> str:
        """Extracts the main content from a PDF file."""
        try:
            documents = PDFReader().read(url)
            content = [document.content for document in documents]
            content = "\n\n".join(content)
            return content
        except:
            return ""

    def crawl(self, url: str, starting_depth: int = 1) -> Dict[str, str]:
        """Crawls an url and returns a dictionary of URLs and their corresponding content."""

        num_links = 0
        crawler_result: Dict[str, str] = {}
        primary_domain = self._get_primary_domain(url)

        self._urls_to_crawl.append((url, starting_depth))

        headers = {
            'accept': '*/*',
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.64 Safari/537.36 Edg/101.0.1210.53',
            'Accept-Language': 'en-US,en;q=0.9,it;q=0.8,es;q=0.7',
            'referer': 'https://www.google.com/',
        }

        with httpx.Client() as client:

            while self._urls_to_crawl:

                # Unpack URL and depth from the global list
                current_url, current_depth = self._urls_to_crawl.pop(0)

                # Skip if...
                if (
                    current_url in self._visited
                    or not urlparse(current_url).netloc.endswith(primary_domain)
                    or current_depth > self.max_depth
                    or num_links >= self.max_links
                ):
                    continue

                self._visited.add(current_url)

                # Delay
                if self.delay:
                    self.sleep()

                # Crawler
                try:
                    logger.debug(f"Reading: {current_url}")
                    response = client.get(current_url, timeout=10, headers=headers)
                    soup = BeautifulSoup(response.content, "html.parser")

                    # Extract main content
                    if current_url.endswith(".pdf"):
                        main_content = self._extract_pdf_content(current_url)
                    else:
                        main_content = self._extract_html_content(soup)

                    crawler_result[current_url] = main_content
                    num_links += 1

                    # Add found URLs to the global list, with incremented depth
                    for link in soup.find_all("a", href=True):
                        full_url = urljoin(current_url, link["href"])
                        parsed_url = urlparse(full_url)

                        # TODO: adjust to use PDFReader and TextReader if files are found
                        if parsed_url.netloc.endswith(primary_domain) and not any(
                            parsed_url.path.endswith(ext) for ext in [".jpg", ".png"]
                        ):
                            if full_url not in self._visited and (full_url, current_depth + 1) not in self._urls_to_crawl:
                                self._urls_to_crawl.append((full_url, current_depth + 1))

                except Exception as e:
                    logger.debug(f"Failed to crawl: {current_url}: {e}")
                    pass

        return crawler_result

    def read(self, url: str) -> List[Document]:
        """Reads a website and returns a list of documents."""

        documents = []

        result = self.crawl(url)

        for url, content in result.items():
            documents.append(
                Document(
                    id=self.key(str(url)),
                    content=content,
                    metadata={"url": str(url)},
                )
            )
        
        return documents


from abc import ABC, abstractmethod
from typing import Any, Callable, Type
from pydantic import BaseModel, Field, model_validator, ConfigDict

from langchain_core.tools import StructuredTool
from langchain_core.utils.function_calling import convert_to_openai_tool

from gwenflow.tools.utils import function_to_json


class BaseTool(BaseModel, ABC):

    name: str
    """The unique name of the tool that clearly communicates its purpose."""

    description: str
    """Used to tell the model how to use the tool."""

    openai_schema: dict = None
    """OpenAI JSON schema"""

    tool_type: str = "function"
    """Tool type: function, langchain, llamaindex."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.openai_schema:
            self.openai_schema = function_to_json(self._run, name=self.name, description=self.description)

    @abstractmethod
    def _run(self, **kwargs: Any) -> Any:
        """Actual implementation of the tool."""

    def run(self, **kwargs: Any) -> Any:
        return self._run(**kwargs)
    
    

class Tool(BaseTool):

    func: Callable
    """The function that will be executed when the tool is called."""

    def _run(self, **kwargs: Any) -> Any:
        if self.tool_type == "langchain":
            return self.func(kwargs)
        return self.func(**kwargs)

    @classmethod
    def from_function(cls, func: Callable) -> "Tool":
        if cls == Tool:
            def _make_with_name(tool_name: str) -> Callable:
                def _make_tool(f: Callable) -> Tool:
                    if f.__doc__ is None:
                        raise ValueError("Function must have a docstring")
                    if f.__annotations__ is None:
                        raise ValueError("Function must have type annotations")
                    return Tool(
                        name=tool_name,
                        description=f.__doc__,
                        func=f,
                        openai_schema=function_to_json(f),
                        tool_type="function",
                    )
                return _make_tool

            if callable(func):
                return _make_with_name(func.__name__)(func)

            if isinstance(func, str):
                return _make_with_name(func)

        raise ValueError(f"Invalid arguments for {cls.__name__}")

    @classmethod
    def from_langchain(cls, tool: StructuredTool) -> "Tool":
        if cls == Tool:
            if tool.run is None:
                raise ValueError("StructuredTool must have a callable 'func'")
            return Tool(
                name=tool.name,
                description=tool.description,
                openai_schema=convert_to_openai_tool(tool),
                func=tool.run,
                tool_type="langchain",
            )
        raise NotImplementedError(f"from_langchain not implemented for {cls.__name__}")


import json
from collections import defaultdict, deque
from pydantic import BaseModel, Field, field_validator
from gwenflow.flows import Flow
from gwenflow.tools import WikipediaTool, WebsiteTool, DuckDuckGoSearchTool
from jinja2 import Environment, FileSystemLoader
from markupsafe import Markup

HTML_TEMPLATE = """
<div>
  <div class="title-box">
    <i class="fas fa-robot"></i>&nbsp;&nbsp;{name}
  </div>
  <div class="box">
    <p>{role}</p>
  </div>
</div>
"""

def escapejs(value):
    """
    Escapes special characters for JavaScript.
    """
    if value is None:
        return ''
    escaped = (
        value.replace('\\', '\\\\')
             .replace("'", "\\'")
             .replace('"', '\\"')
             .replace('\n', '\\n')
             .replace('\r', '\\r')
             .replace('\t', '\\t')
    )
    return Markup(escaped)

class Node(BaseModel):
    id: str
    name: str
    context_vars: list[str] = Field(default_factory=list)
    role: str
    inputs: dict = Field(default_factory=lambda: {"input_1": {"connections": []}})
    outputs: dict = Field(default_factory=lambda: {"output_1": {"connections": []}})
    pos_x: int = 0
    pos_y: int = 0

    @field_validator("id", "name", "role", mode="before")
    def non_empty_string(cls, value):
        if not value.strip():
            raise ValueError("Field cannot be an empty string.")
        return value

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "context_vars": self.context_vars,
            "class": "agent",
            "html": HTML_TEMPLATE.format(name=self.name, role=self.role),
            "typenode": False,
            "data": {},
            "inputs": self.inputs,
            "outputs": self.outputs,
            "pos_x": self.pos_x,
            "pos_y": self.pos_y,
        }

class DrawFlow(BaseModel):
    flow: Flow
    nodes: dict[str, Node] = Field(default_factory=dict)

    def _initialize_nodes(self):
        """Creates Node objects from the flow agents."""
        self.nodes = {
            str(agent.id): Node(
                id=str(agent.id),
                name=agent.name,
                context_vars=agent.context_vars,
                role=agent.role
            )
            for agent in self.flow.__dict__["agents"]
        }

    def _find_longest_path_to_sources(self, start_node):
        """Finds the longest path to sources in a graph. Based on BFS Algorithm."""
        adj_list = defaultdict(list)
        in_degree = {node_id: 0 for node_id in self.nodes}

        for node in self.nodes.values():
            if not node.outputs:
                continue
            for conn in node.outputs["output_1"]["connections"]:
                adj_list[node.id].append(conn["node"])
                adj_list[conn["node"]].append(node.id)
                in_degree[conn["node"]] += 1

        sources = [node_id for node_id, degree in in_degree.items() if degree == 0]
        distances = {node_id: float('inf') for node_id in self.nodes}
        distances[start_node] = 0

        queue = deque([start_node])
        while queue:
            current = queue.popleft()
            for neighbor in adj_list[current]:
                if distances[neighbor] > distances[current] + 1:
                    distances[neighbor] = distances[current] + 1
                    queue.append(neighbor)

        return min((distances[source] for source in sources), default=float('inf'))

    def _assign_io_connections(self):
        """Assigns input and output connections between nodes based on context variables."""
        for node in self.nodes.values():
            for other_node in self.nodes.values():
                if node.id == other_node.id:
                    continue
                if node.name in other_node.context_vars:
                    node.outputs["output_1"]["connections"].append({
                        "node": other_node.id,
                        "output": "input_1"
                    })
                    other_node.inputs["input_1"]["connections"].append({
                        "node": node.id,
                        "input": "output_1"
                    })
        for node in self.nodes.values():
            if not node.inputs:
                continue
            if len(node.inputs["input_1"]["connections"]) == 0:
                node.inputs = {}
            if not node.outputs:
                continue
            if len(node.outputs["output_1"]["connections"]) == 0:
                node.outputs = {}

    def _assign_positions(self, separation_x=400, separation_y=200, initial_x=-350, initial_y=-150):
        """Calculates and assigns positions to nodes."""
        columns = {node_id: self._find_longest_path_to_sources(node_id) + 1 for node_id in self.nodes}
        rows = defaultdict(int)
        columns_inv = defaultdict(list)

        for node_id, col in columns.items():
            columns_inv[col].append(node_id)

        for col, nodes in columns_inv.items():
            for index, node_id in enumerate(nodes):
                rows[node_id] = index + 1

        for node_id, node in self.nodes.items():
            node.pos_x = initial_x + columns[node_id] * separation_x
            node.pos_y = initial_y + rows[node_id] * separation_y

    def generate_drawflow(self):
        """Generates the drawflow JSON structure."""
        self._assign_io_connections()
        self._assign_positions()
        return {
            "drawflow": {
                "Home": {
                    "data": {node_id: node.to_dict() for node_id, node in self.nodes.items()}
                }
            }
        }

    @classmethod
    def from_yaml(cls, yaml_path, tools):
        """Creates a DrawFlow instance from a YAML configuration."""
        flow = Flow().from_yaml(yaml_path, tools=tools)
        instance = cls(flow=flow)
        instance._initialize_nodes()
        return instance
    
    @classmethod
    def from_flow(cls, flow):
        """Creates a DrawFlow instance from a Flow instance."""
        instance = cls(flow=flow)
        instance._initialize_nodes()
        return instance

    def save_to_json(self, json_path):
        """Saves the generated drawflow to a JSON file."""
        drawflow_json = self.generate_drawflow()
        with open(json_path, "w", encoding="utf-8") as json_file:
            json.dump(drawflow_json, json_file, indent=4, ensure_ascii=False)

    def generate_html(self, output_file):
        """Generates an HTML file using a Jinja2 template."""
        template_file = "index.html"
        drawflow_json = self.generate_drawflow()
        env = Environment(loader=FileSystemLoader(""))
        env.filters['escapejs'] = escapejs

        template = env.get_template(template_file)

        rendered_html = template.render({
            "drawflow_json": json.dumps(drawflow_json)
        })

        with open(output_file, "w", encoding="utf-8") as f:
            f.write(rendered_html)

        print(f"HTML generated successfully at {output_file}")


from typing import Any
import logging
import boto3
from opensearchpy import OpenSearch, helpers, AWSV4SignerAuth, RequestsHttpConnection

import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


logger = logging.getLogger(__name__)


class OpenSearchDocumentStore():

    def __init__(self, uri, index, use_ssl=True, verify_certs=False, ca_certs=None, timeout=30, aws=False):
        if aws:
            credentials = boto3.Session().get_credentials()
            auth = AWSV4SignerAuth(credentials, "eu-west-3", "es")
            self._client = OpenSearch(
                hosts = [{'host': uri, "port": 443}],
                http_auth = auth,
                use_ssl = True,
                verify_certs = True,
                connection_class = RequestsHttpConnection,
                pool_maxsize = 20
            )
        else:
            self._client = OpenSearch(uri, http_compress=True, use_ssl=use_ssl, verify_certs=verify_certs, ca_certs=ca_certs, ssl_assert_hostname=False, ssl_show_warn=False)
        self._index = index
        self._timeout = timeout

    @property
    def client(self) -> Any:
        return self._client

    @property
    def index(self) -> str:
        return self._index

    def count(self, query: str):
        q = { "query": query, "track_total_hits": True, "size": 1 }
        resp = self._client.search(body=q, index=self._index, request_timeout=self._timeout)
        return resp['hits']['total']['value']

    def delete(self, ids=None):
        if not ids:
            return self._client.indices.delete(index=self._index)
        if not isinstance(ids, list):
            ids = [ids]
        if len(ids)==0:
            logger.warning("No documents to be deleted from OpenSearch.")
            return False
        for id in ids:
            self._client.delete(index=self._index, id=id)
        return True

    def get(self, id: str):
        try:
            data = self._client.get(id=id, index=self._index)
            if "_source" in data:
                return data["_source"]
        except Exception as e:
            logger.error(repr(e))
        return None

    def put(self, documents, column_id="id"):
        if not isinstance(documents, list):
            documents = [documents]
        if len(documents)==0:
            logger.warning("No documents to be added to OpenSearch.")
            return False
        def _generator(documents):
            for doc in documents:
                yield {"_index" : self._index, "_id": doc[column_id], "_source": doc}
        success, failed = helpers.bulk(self._client, _generator(documents))
        if success < len(documents):
            logger.warning(f"Some documents failed to be added to OpenSearch. Failures: {failed}")
        logger.info(f"Indexed {success} documents")
        return True

    def search(self, query, aggs=None, sort=None, page=1, per_page=5000, fields=None, all=False, track_total_hits=True):

        # if all==True:
        #     per_page = 10000

        body = { "query": query, "size": per_page }

        if (page > 1) and (all == False):
            body["from"] = (page-1) * per_page

        if track_total_hits:
            body["track_total_hits"] = True

        if fields:
            body["fields"] = fields

        if aggs:
            body["aggs"]=aggs

        if sort:
            body["sort"] = sort

        # réfléchir au search after pour remplacer le scroll
        # if all:
        #     q["search_after"] = search_after
        
        documents = []
        aggregations = []
        if not all:
            resp = self._client.search(body=body, index=self._index, request_timeout=self._timeout)
            total = resp['hits']['total']['value']
            documents = [hit["_source"] for hit in resp['hits']['hits']]
            if aggs:
                aggregations = resp['aggregations']
            return {"object": "list", "total": total, "page": page, "per_page": per_page, "data": documents, "aggs":aggregations}

        resp = self._client.search(body=body, index=self._index, request_timeout=self._timeout, scroll='1m')
        for doc in resp['hits']['hits']:
            documents.append(doc["_source"])
        scroll_id = resp['_scroll_id']
        while len(resp['hits']['hits']):
            try:
                resp = self._client.scroll(scroll_id=scroll_id, scroll='5m')
                for doc in resp['hits']['hits']:
                    documents.append(doc["_source"])
                if scroll_id != resp["_scroll_id"]:
                    self._client.clear_scroll(scroll_id=scroll_id)
                    scroll_id = resp['_scroll_id']
            except Exception as e:
                logger.error(repr(e))
                break
        self._client.clear_scroll(scroll_id=scroll_id)
        total = len(documents)
        per_page = len(documents)
        return {"object": "list", "total": total, "page": page, "per_page": per_page, "data": documents, "aggs":aggregations}



from typing import Tuple

import re
import dirtyjson as json

from gwenflow.agents.react.types import ActionReasoningStep
from gwenflow.utils import extract_json_str


def extract_tool_use(input_text: str) -> Tuple[str, str, str]:
    pattern = (
        r"\s*Thought: (.*?)\n+Action: ([^\n\(\) ]+).*?\n+Action Input: .*?(\{.*\})"
    )

    match = re.search(pattern, input_text, re.DOTALL)
    if not match:
        raise ValueError(f"Could not extract tool use from input text: {input_text}")

    thought = match.group(1).strip()
    action = match.group(2).strip()
    action_input = match.group(3).strip()
    return thought, action, action_input

def action_input_parser(json_str: str) -> dict:
    processed_string = re.sub(r"(?<!\w)\'|\'(?!\w)", '"', json_str)
    pattern = r'"(\w+)":\s*"([^"]*)"'
    matches = re.findall(pattern, processed_string)
    return dict(matches)

def extract_final_response(input_text: str) -> Tuple[str, str]:
    pattern = r"\s*Thought:(.*?)Final Answer:(.*?)(?:$)"

    match = re.search(pattern, input_text, re.DOTALL)
    if not match:
        raise ValueError(
            f"Could not extract final answer from input text: {input_text}"
        )

    thought = match.group(1).strip()
    answer = match.group(2).strip()
    return thought, answer

def parse_action_reasoning_step(output: str) -> ActionReasoningStep:
    # Weaker LLMs may generate ReActAgent steps whose Action Input are horrible JSON strings.
    # `dirtyjson` is more lenient than `json` in parsing JSON strings.
    thought, action, action_input = extract_tool_use(output)
    json_str = extract_json_str(action_input)
    # First we try json, if this fails we use ast
    try:
        action_input_dict = json.loads(json_str)
    except Exception:
        action_input_dict = action_input_parser(json_str)
    return ActionReasoningStep(
        thought=thought, action=action, action_input=action_input_dict
    )

def extract_thought(input_text: str) -> str:
    pattern = r"\s*Thought:(.*?)$"
    match = re.search(pattern, input_text, re.DOTALL)
    if not match:
        raise ValueError(f"Could not extract tought from input text: {input_text}")
    return match.group(1).strip()

class ReActOutputParser:
    """ReAct Output parser."""

    def parse(self, output: str) -> ActionReasoningStep:

        # Agent directly outputs the answer instead of following the thought-answer format
        if "Thought:" not in output:
            return ActionReasoningStep(
                thought="I can answer without any more tools!",
                response=output,
            )

        # An "Action" should take priority over an "Answer"
        if "Action:" in output:
            return parse_action_reasoning_step(output)

        if "Final Answer:" in output:
            thought, answer = extract_final_response(output)
            return ActionReasoningStep(thought=thought, response=answer, is_done=True)

        if "Thought:" in output:
            thought = extract_thought(output)
            return ActionReasoningStep(thought=thought)
        
        return ActionReasoningStep(thought=output)



import json
from typing import List, Dict, Any, Optional
from pydantic import BaseModel, model_validator, field_validator, Field, ConfigDict

from gwenflow.tools import BaseTool
from gwenflow.types.document import Document
from gwenflow.knowledge.base import Knowledge
from gwenflow.utils import logger


class KnowledgeTool(BaseTool):

    name: str = "KnowledgeTool"
    description: str = "Search in the knowledge base."

    knowledge: Optional[Knowledge] = Field(None, validate_default=True)


    @field_validator("knowledge", mode="before")
    @classmethod
    def set_knowledge(cls, v: Optional[Knowledge]) -> Knowledge:
        if not v:
            try:
                v = Knowledge()
            except Exception as e:
                logger.error(f"Error creating KnowledgeTool: {e}")
        return v

    def load_documents(self, documents: List[Any]) -> bool:
        for document in documents:
            if isinstance(document, str):
                document = Document(content=document)
            self.knowledge.load_document(document)
    
    def _run(self, query: str = Field(description="The search query.")) -> str:
        documents = self.knowledge.search(query, limit=5)
        return json.dumps([doc.to_dict() for doc in documents])




import uuid
import json
import inspect
from typing import List, Union, Optional, Any, Dict, Iterator
from collections import defaultdict
from pydantic import BaseModel, model_validator, field_validator, Field, UUID4
from datetime import datetime

from gwenflow.llms import ChatOpenAI
from gwenflow.types import ChatCompletionMessage, ChatCompletionMessageToolCall
from gwenflow.tools import BaseTool
from gwenflow.memory import ChatMemoryBuffer
from gwenflow.agents.types import AgentResponse
from gwenflow.agents.utils import merge_chunk
from gwenflow.agents.prompts import PROMPT_TOOLS, PROMPT_STEPS, PROMPT_TOOLS_REACT_GUIDELINES, PROMPT_TASK
from gwenflow.utils import logger


MAX_TURNS = float('inf') #10


class Agent(BaseModel):

    # --- Agent Settings
    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    name: str = Field(description="Name of the agent")
    role: str = Field(description="Role of the agent")
    description: Optional[str] = Field(default=None, description="Description of the agent")

    # --- Settings for system message
    system_prompt: Optional[str] = None
    instructions: Optional[Union[str, List[str]]] = []
    add_datetime_to_instructions: bool = True
    markdown: bool = False
    response_model: Optional[dict] = None
    steps: Optional[List[str]] = []
    follow_steps: bool = False
    is_react: bool = False
    system_prompt_allowed: bool = True
 
    # --- Agent Model and Tools
    llm: Optional[Any] = Field(None, validate_default=True)
    tools: List[BaseTool] = []
    tool_choice: Optional[Union[str, Dict[str, Any]]] = None
    show_tool_calls: bool = False

    # --- Reasoning models
    reasoning_model: Optional[Any] = Field(None, validate_default=True)

    # --- Task, Context and Memory
    context_vars: Optional[List[str]] = []
    memory: Optional[ChatMemoryBuffer] = None
    metadata: Optional[Dict[str, Any]] = None
    # knowledge: Optional[Knowledge] = None

    # --- Team of agents
    team: Optional[List["Agent"]] = None


    @field_validator("id", mode="before")
    @classmethod
    def deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise ValueError("This field is not to be set by the user.")

    @field_validator("instructions", mode="before")
    @classmethod
    def set_instructions(cls, v: Optional[Union[List, str]]) -> str:
        if isinstance(v, str):
            instructions = [v]
            return instructions
        return v

    @field_validator("llm", mode="before")
    @classmethod
    def set_llm(cls, v: Optional[Any]) -> Any:
        llm = v or ChatOpenAI(model="gpt-4o-mini")
        return llm

    @model_validator(mode="after")
    def model_valid(self) -> Any:
        if self.memory is None and self.llm is not None:
             token_limit = self.llm.get_context_window_size()
             self.memory = ChatMemoryBuffer(token_limit=token_limit)
        return self
    
    def get_system_message(self, context: Optional[Any] = None):
        """Return the system message for the Agent."""

        if self.system_prompt is not None:
            return dict(role="system", content=self.system_prompt)

        system_message_lines = []

        # name, role and description
        txt = f"You are an AI agent named '{self.name}'."
        if self.role:
            txt += f" Your role is: {self.role.strip('.')}."
        if self.description:
            txt += f" {self.description.strip('.')}."
        system_message_lines.append(f"{txt}\n")

        if self.add_datetime_to_instructions:
            txt = f"The current date and time is:\n<current_date>\n{ datetime.now() }\n</current_date>\n"
            system_message_lines.append(txt)

        # tools
        if self.tools:
            if self.is_react:
                tools = self.get_tools_text_schema()
                system_message_lines.append(PROMPT_TOOLS.format(tools=tools).strip())
                system_message_lines.append(PROMPT_TOOLS_REACT_GUIDELINES)
                system_message_lines.append("")
            else:
                tools = ",".join(self.get_tool_names())
                system_message_lines.append(PROMPT_TOOLS.format(tools=tools).strip())
                system_message_lines.append("")

        # instructions
        instructions = []        
        if self.response_model:
            instructions.append("Use JSON to format your answers.")
        elif self.markdown:
            instructions.append("Use markdown to format your answers.")
        if self.tools is not None:
            instructions.append("Only use the tools you are provided.")
        if context is not None:
            instructions.append("Always prefer information from the provided context over your own knowledge.")

        if len(self.instructions) > 0:
            instructions += self.instructions

        if len(instructions) > 0:
            system_message_lines.append("## Guidelines:")
            system_message_lines.extend([f"- {instruction}" for instruction in instructions])
            system_message_lines.append("")

        if self.response_model:
            system_message_lines.append("## Provide your output using the following JSON schema:")
            system_message_lines.append("<json_schema>")
            system_message_lines.append(json.dumps(self.response_model, indent=4))
            system_message_lines.append("</json_schema>")
            system_message_lines.append("")

        if self.follow_steps:
            if len(self.steps) > 0:
                system_message_lines.append("## Follow these steps:\n")
                system_message_lines.extend([f"{i+1}. {step}" for i, step in enumerate(self.steps)])
                system_message_lines.append("")
            else:
                system_message_lines.append(PROMPT_STEPS.strip())
                system_message_lines.append("")

        # final system prompt
        if len(system_message_lines) > 0:
            return dict(role="system", content=("\n".join(system_message_lines)).strip())
        
        return None


    def format_context(self, context):
        prompt  = "Use the following information if it helps:\n\n"

        if isinstance(context, str):
            prompt += "<context>\n"
            prompt += context + "\n"
            prompt += "</context>\n\n"

        elif isinstance(context, dict):
            for key in context.keys():
                prompt += f"<{key}>\n"
                prompt += context.get(key) + "\n"
                prompt += f"</{key}>\n\n"

        return prompt

    def get_user_message(self, task: Optional[str] = None, context: Optional[Any] = None):
        """Return the user message for the Agent."""

        if not task and not context:
            raise ValueError("You need a task or a context (or both) to run the agent!")

        prompt = ""

        if context:
            prompt += self.format_context(context)

        if task:
            prompt += PROMPT_TASK.format(task=task)

        return { "role": "user", "content": prompt }

    
    def get_tools_openai_schema(self):
        return [tool.openai_schema for tool in self.tools]

    def get_tools_text_schema(self) -> str:
        descriptions = []
        for tool in self.tools:
            sig = inspect.signature(tool._run)
            description = f"{tool.name}{sig} - {tool.description}"
            descriptions.append(description)
        return "\n".join(descriptions)

    def get_tools_map(self):
        return {tool.name: tool for tool in self.tools}

    def get_tool_names(self):
        return [tool.name for tool in self.tools]

    def execute_tool_call(self, tool_name: str, arguments: Dict[str, str]) -> Any:

        available_tools = self.get_tools_map()
        if tool_name not in available_tools:
            logger.error(f"Unknown tool {tool_name}, should be instead one of { available_tools.keys() }.")
            return None

        logger.debug(f"Tool call: {tool_name}({arguments})")
        observation = available_tools[tool_name].run(**arguments)

        return observation

    def handle_tool_calls(
        self,
        tool_calls: List[ChatCompletionMessageToolCall],
    ) -> List:
        
        tool_map = self.get_tools_map()

        messages = []
        for tool_call in tool_calls:

            tool_name = tool_call.function.name

            # handle missing tool case, skip to next tool
            if tool_name not in tool_map.keys():
                logger.warning(f"Unknown tool {tool_name}, should be instead one of { tool_map.keys() }.")
                messages.append(
                    {
                        "role": "tool",
                        "tool_call_id": tool_call.id,
                        "tool_name": tool_name,
                        "content": f"Observation: Error, Tool {tool_name} not found.",
                    }
                )
                continue

            arguments = json.loads(tool_call.function.arguments)
            observation = self.execute_tool_call(tool_name, arguments)
            
            if observation:
                messages.append(
                    {
                        "role": "tool",
                        "tool_call_id": tool_call.id,
                        "tool_name": tool_name,
                        "content": f"Observation:\n{observation}",
                    }
                )

        return messages

    def invoke(self, messages: list, stream: bool = False) ->  Union[Any, Iterator[Any]]:

        tools = self.get_tools_openai_schema()

        params = {
            "messages": messages,
            "tools": tools or None,
            "tool_choice": self.tool_choice,
            "parse_response": False,
        }

        response_format = None
        if self.response_model:
            response_format = {"type": "json_object"}

        if stream:
            return self.llm.stream(**params, response_format=response_format)
        
        return self.llm.invoke(**params, response_format=response_format)

    def _run(
        self,
        task: Optional[str] = None,
        *,
        context: Optional[Any] = None,
        stream: Optional[bool] = False,
    ) ->  Iterator[AgentResponse]:

        # add reasoning
        if self.reasoning_model:
            reasoning_message = self.reason(task=task)
            if reasoning_message:
                if not context:
                    context = {}
                if isinstance(context, str):
                    text = context
                    context = { "context": text }
                context["thinking"] = reasoning_message

        messages_for_model = []

        # system messages
        system_message = self.get_system_message(context=context)
        if system_message:
            messages_for_model.append(system_message)

        # user messages
        user_message = self.get_user_message(task=task, context=context)
        if user_message:
            messages_for_model.append(user_message)
            self.memory.add_message(user_message)

        # global loop
        init_len = len(messages_for_model)
        while len(messages_for_model) - init_len < MAX_TURNS:

            if stream:
                message = {
                    "content": "",
                    "sender": self.name,
                    "role": "assistant",
                    "function_call": None,
                    "tool_calls": defaultdict(
                        lambda: {
                            "function": {"arguments": "", "name": ""},
                            "id": "",
                            "type": "",
                        }
                    ),
                }

                completion = self.invoke(messages=messages_for_model, stream=True)

                for chunk in completion:
                    if len(chunk.choices) > 0:
                        delta = json.loads(chunk.choices[0].delta.json())
                        if delta["role"] == "assistant":
                            delta["sender"] = self.name
                        if delta["content"]:
                            yield AgentResponse(
                                delta=delta["content"],
                                messages=None,
                                agent=self,
                                tools=self.tools,
                            )
                        elif delta["tool_calls"] and self.show_tool_calls:
                            if delta["tool_calls"][0]["function"]["name"] and not delta["tool_calls"][0]["function"]["arguments"]:
                                response = f"""**Calling:** {delta["tool_calls"][0]["function"]["name"]}"""
                                yield AgentResponse(
                                    delta=response,
                                    messages=None,
                                    agent=self,
                                    tools=self.tools,
                                )
                        delta.pop("role", None)
                        delta.pop("sender", None)
                        merge_chunk(message, delta)

                message["tool_calls"] = list(message.get("tool_calls", {}).values())
                message = ChatCompletionMessage(**message)
            
            else:
                completion = self.invoke(messages=messages_for_model)
                message = completion.choices[0].message
                message.sender = self.name

            # add messages to the current message stack
            message_dict = json.loads(message.model_dump_json())
            messages_for_model.append(message_dict)

            if not message.tool_calls:
                logger.debug("Task done.")
                self.memory.add_message(message_dict) # We only keep the answer in memory (not tool calls)
                break

            # handle tool calls and switching agents
            tool_response_messages = self.handle_tool_calls(message.tool_calls)
            messages_for_model.extend(tool_response_messages)

        content = messages_for_model[-1]["content"]
        if self.response_model:
            content = json.loads(content)

        yield AgentResponse(
            content=content,
            messages=messages_for_model[init_len:],
            agent=self,
            tools=self.tools,
        )


    def run(
        self,
        task: Optional[str] = None,
        *,
        context: Optional[Any] = None,
        stream: Optional[bool] = False,
        output_file: Optional[str] = None,
    ) ->  Union[AgentResponse, Iterator[AgentResponse]]:

        logger.debug("")
        logger.debug("------------------------------------------")
        logger.debug(f"Running Agent: { self.name }")
        logger.debug("------------------------------------------")
        logger.debug("")

        if stream:
            response = self._run(
                task=task,
                context=context,
                stream=True,
            )
            return response
    
        else:

            response = self._run(
                task=task,
                context=context,
                stream=False,
            )
            response = next(response)

            if output_file:
                with open(output_file, "a") as file:
                    file.write("\n")
                    file.write("---\n\n")
                    file.write(f"# Agent: { self.name }\n")
                    file.write(f"{ task }\n")
                    file.write("\n")
                    file.write(response.content)

            return response



from gwenflow.readers.directory import SimpleDirectoryReader
from gwenflow.readers.text import TextReader
from gwenflow.readers.json import JSONReader
from gwenflow.readers.pdf import PDFReader
from gwenflow.readers.website import WebsiteReader

__all__ = [
    "SimpleDirectoryReader",
    "TextReader",
    "JSONReader",
    "PDFReader",
    "WebsiteReader",
]


from typing import Any

from pydantic import Field

from gwenflow.tools import BaseTool


class YahooFinanceSearch(BaseTool):
    name: str = "YahooFinanceSearch"
    description: str = "Search for a stock on Yahoo Finance."

    def _run(self, query: str = Field(description="The stock to search for.")) -> str:
        try:
            import yfinance as yf

            stocks = yf.Search(query)

            print(stocks.quotes)
        except ImportError:
            return "yfinance is not installed. Please install it with `pip install yfinance`."


class YahooFinancePick(BaseTool):
    name: str = "YahooFinancePick"
    description: str = "Search for a stock on Yahoo Finance."

    def _run(self, query: str = Field(description="The stock to search for.")) -> str:
        try:
            import yfinance as yf

            stocks = yf.Search(query)

            print(stocks.quotes)
        except ImportError:
            return "yfinance is not installed. Please install it with `pip install yfinance`."


class YahooFinanceStock(BaseTool):
    name: str = "YahooFinanceStock"
    description: str = "Retrieve stock data from Yahoo Finance."

    def _run(
        self, ticker: str = Field(description="The ticker stock to search for.")
    ) -> dict | str:
        try:
            import yfinance as yf

            stock = yf.Ticker(ticker)

            return stock.info

        except ImportError:
            return "yfinance is not installed. Please install it with `pip install yfinance`."


class YahooFinanceNews(BaseTool):
    name: str = "YahooFinanceNews"
    description: str = (
        "Search for news on company on Yahoo Finance."
        "Search for news on Yahoo Finance."
        "Input should be a search query."
        "This tool will return the latest news"
    )

    def _run(
        self, query: str = Field(description="The query to search for.")
    ) -> dict[Any, Any] | str:
        try:
            import yfinance as yf

            stocks = yf.Search(
                query, enable_fuzzy_query=True, news_count=15, max_results=0
            )
            # create a dict of news with title and link
            news_dict = {}
            for news in stocks.news:
                news_dict[news["title"]] = news["link"]

            return news_dict
        except ImportError:
            return "yfinance is not installed. Please install it with `pip install yfinance`."


class YahooFinanceScreen(BaseTool):
    name: str = "YahooFinanceScreen"
    description: str = (
        "Screen for stocks on Yahoo Finance."
        "This tool will return a list of stocks that meet the criteria."
    )

    def _run(
        self,
        operator: str = Field(
            description="The queries support operators: GT (greater than), LT (less than), BTWN (between), EQ (equals), and logical operators AND and OR for combining multiple conditions."
        ),
        values: list[str | int] = Field(
            description="The values to search for. For example, ['intradaymarketcap', 1000000000] for market cap greater than 1 billion.",
        ),
    ) -> list[Any] | str:
        try:
            import yfinance as yf

            screener = yf.Screener()

            try:

                query = yf.EquityQuery(operator, eval(values))

            except Exception:
                return "Query not valid. Please check the Yahoo Finance screener documentation."

            print(query.to_dict())

            try:
                screener.set_body(
                    {
                        "size": 250,
                        "offset": 0,
                        "sortField": "avgdailyvol3m",
                        "sortType": "DESC",
                        "quoteType": "EQUITY",
                        "query": query.to_dict(),
                        "userId": "",
                        "userIdType": "guid",
                    }
                )

                result = screener.response
                symbols = [quote["symbol"] for quote in result["quotes"]]

                print(symbols)

                return symbols
            except Exception:
                return "Query not valid. Please check the Yahoo Finance screener documentation."

        except ImportError:
            return "yfinance is not installed. Please install it with `pip install yfinance`."



import json
from pydantic import Field

from gwenflow.tools import BaseTool
from gwenflow.readers.website import WebsiteReader


class WebsiteTool(BaseTool):

    name: str = "WebsiteTool"
    description: str = "Fetches and returns the content of a given URL."

    def _run(self, url: str = Field(description="The url of the website to read.")) -> str:
        reader = WebsiteReader(max_depth=1)
        documents = reader.read(url)
        return json.dumps([doc.to_dict() for doc in documents])



from typing import List, Optional

from pydantic import BaseModel, ConfigDict
from gwenflow.types import Document



class Reranker(BaseModel):
    """Base class for rerankers"""

    model: str
    top_k: Optional[int] = None
    threshold: Optional[float] = None

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    def rerank(self, query: str, documents: List[Document]) -> List[Document]:
        raise NotImplementedError
    


import os
from typing import Optional, Dict, Any

from openai import AzureOpenAI, AsyncAzureOpenAI

from gwenflow.llms.openai import ChatOpenAI


class ChatAzureOpenAI(ChatOpenAI):
 
    api_version: Optional[str] = None
    azure_endpoint: Optional[str] = None
    azure_deployment: Optional[str] = None

    def _get_client_params(self) -> Dict[str, Any]:

        api_key = self.api_key or os.environ.get("AZURE_OPENAI_API_KEY")
        azure_endpoint = self.azure_endpoint or os.environ.get("AZURE_OPENAI_ENDPOINT")
        azure_deployment = self.azure_deployment or os.environ.get("AZURE_OPENAI_DEPLOYMENT")
        api_version = self.api_version or os.environ.get("AZURE_OPENAI_API_VERSION")

        if api_key is None:
            raise ValueError(
                "The api_key client option must be set either by passing api_key to the client or by setting the AZURE_OPENAI_API_KEY environment variable"
            )

        client_params = {
            "api_key": api_key,
            "api_version": api_version,
            "azure_endpoint": azure_endpoint,
            "azure_deployment": azure_deployment,
            "timeout": self.timeout,
            "max_retries": self.max_retries,
        }
        
        client_params = {k: v for k, v in client_params.items() if v is not None}
        
        return client_params

    def get_client(self) -> AzureOpenAI:

        if self.client:
            return self.client
        
        client_params = self._get_client_params()

        self.client = AzureOpenAI(**client_params)
        return self.client

    def get_async_client(self) -> AsyncAzureOpenAI:

        if self.async_client:
            return self.async_client
        
        client_params = self._get_client_params()

        self.async_client = AsyncAzureOpenAI(**client_params)
        return self.async_client
    



from pydantic import BaseModel, Field
from typing import Optional, Union, List, Any, Dict, Generator
from typing_extensions import Literal



class CompletionUsage(BaseModel):
    prompt_tokens: int
    """The number of tokens used by the prompt."""

    completion_tokens: int
    """Number of tokens in the generated completion."""

    total_tokens: int
    """The total number of tokens used by the request."""


class Function(BaseModel):
    arguments: str
    name: str
    """The name of the function to call."""


class ChatCompletionMessageToolCall(BaseModel):
    id: str
    """The ID of the tool call."""

    function: Function
    """The function that the model called."""

    type: Literal["function"]
    """The type of the tool. Currently, only `function` is supported."""


class ChatCompletionAudio(BaseModel):
    id: str
    """Unique identifier for this audio response."""

    data: str
    """
    Base64 encoded audio bytes generated by the model, in the format specified in
    the request.
    """

    expires_at: int
    """
    The Unix timestamp (in seconds) for when this audio response will no longer be
    accessible on the server for use in multi-turn conversations.
    """

    transcript: str
    """Transcript of the audio generated by the model."""


class ChatCompletionMessage(BaseModel):
    content: Optional[str] = None
    """The contents of the message."""

    refusal: Optional[str] = None
    """The refusal message generated by the model."""

    role: Literal["assistant"]
    """The role of the author of this message."""

    audio: Optional[ChatCompletionAudio] = None
    """
    If the audio output modality is requested, this object contains data about the
    audio response from the model.
    """

    tool_calls: Optional[List[ChatCompletionMessageToolCall]] = None
    """The tool calls generated by the model, such as function calls."""



# class ChoiceLogprobs(BaseModel):
#     content: Optional[List[ChatCompletionTokenLogprob]] = None
#     """A list of message content tokens with log probability information."""

#     refusal: Optional[List[ChatCompletionTokenLogprob]] = None
#     """A list of message refusal tokens with log probability information."""


class Choice(BaseModel):
    finish_reason: Literal["stop", "length", "tool_calls", "content_filter", "function_call"]
    """The reason the model stopped generating tokens.

    This will be `stop` if the model hit a natural stop point or a provided stop
    sequence, `length` if the maximum number of tokens specified in the request was
    reached, `content_filter` if content was omitted due to a flag from our content
    filters, `tool_calls` if the model called a tool, or `function_call`
    (deprecated) if the model called a function.
    """

    index: int
    """The index of the choice in the list of choices."""

    # logprobs: Optional[ChoiceLogprobs] = None
    # """Log probability information for the choice."""

    message: ChatCompletionMessage
    """A chat completion message generated by the model."""


class ChatCompletion(BaseModel):
    id: str
    """A unique identifier for the chat completion."""

    choices: List[Choice]
    """A list of chat completion choices.

    Can be more than one if `n` is greater than 1.
    """

    created: int
    """The Unix timestamp (in seconds) of when the chat completion was created."""

    model: str
    """The model used for the chat completion."""

    object: Literal["chat.completion"]
    """The object type, which is always `chat.completion`."""

    service_tier: Optional[Literal["scale", "default"]] = None
    """The service tier used for processing the request.

    This field is only included if the `service_tier` parameter is specified in the
    request.
    """

    system_fingerprint: Optional[str] = None
    """This fingerprint represents the backend configuration that the model runs with.

    Can be used in conjunction with the `seed` request parameter to understand when
    backend changes have been made that might impact determinism.
    """

    usage: Optional[CompletionUsage] = None
    """Usage statistics for the completion request."""




import uuid
from typing import Optional, Any, Dict
from pydantic import BaseModel, ConfigDict, field_validator, Field


class Document(BaseModel):
    """Base class for Documents"""

    id: Optional[str] = Field(None, validate_default=True)
    content: str
    metadata: Dict[str, Any] = {}
    score: Optional[float] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("id", mode="before")
    def set_id(cls, v: Optional[str]) -> str:
        id = v or str(uuid.uuid4())
        return id

    def to_dict(self) -> Dict[str, Any]:
        """Returns a dictionary representation of the document"""
        return self.model_dump(include={"id", "content", "metadata", "score"}, exclude_none=True)

    @classmethod
    def from_dict(cls, document: Dict[str, Any]) -> "Document":
        """Returns a Document object from a dictionary representation"""
        return cls.model_validate(**document)

    @classmethod
    def from_json(cls, document: str) -> "Document":
        """Returns a Document object from a json string representation"""
        return cls.model_validate_json(document)


import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__)
except importlib.metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = ""


from gwenflow.llms import ChatGwenlake, ChatOpenAI, ChatAzureOpenAI, ChatOllama
from gwenflow.readers import SimpleDirectoryReader
from gwenflow.agents import Agent, ReActAgent
from gwenflow.tools import Tool
from gwenflow.flows import Flow, AutoFlow
from gwenflow.types import Document
from gwenflow.knowledge import Knowledge


__all__ = [
    "ChatGwenlake",
    "ChatOpenAI",
    "ChatAzureOpenAI",
    "ChatOllama",
    "Document",
    "SimpleDirectoryReader",
    "Agent",
    "ReActAgent",
    "Tool",
    "Flow",
    "AutoFlow",
    "Knowledge",
]


import os
from typing import Dict, Any

from gwenflow.llms.openai import ChatOpenAI


class ChatGwenlake(ChatOpenAI):
 
    model: str
    base_url: str = "https://api.gwenlake.com/v1"

    def _get_client_params(self) -> Dict[str, Any]:

        api_key = self.api_key
        if api_key is None:
            api_key = os.environ.get("GWENLAKE_API_KEY")
        if api_key is None:
            raise ValueError(
                "The api_key client option must be set either by passing api_key to the client or by setting the GWENLAKE_API_KEY environment variable"
            )

        organization = self.organization
        if organization is None:
            organization = os.environ.get('GWENLAKE_ORGANIZATION')

        client_params = {
            "api_key": api_key,
            "organization": organization,
            "base_url": self.base_url,
            "timeout": self.timeout,
            "max_retries": self.max_retries,
        }

        client_params = {k: v for k, v in client_params.items() if v is not None}

        return client_params
        



from typing import List, Optional, Any
from pydantic import BaseModel, Field
from time import time


class AgentResponse(BaseModel):
    content: Optional[Any] = None
    content_type: str = "str"
    delta: Optional[str] = None
    messages: Optional[List] = None
    agent: Optional[Any] = None
    tools: Optional[List[Any]] = None
    created_at: int = Field(default_factory=lambda: int(time()))



from typing import Optional, Union, Any, List, Dict

from gwenflow.llms.openai import ChatOpenAI

class ChatOllama(ChatOpenAI):
 
    base_url: str

    def _get_client_params(self) -> Dict[str, Any]:

        client_params = {
            "base_url": self.base_url,
            "api_key": self.api_key,
            "timeout": self.timeout,
            "max_retries": self.max_retries,
        }

        client_params = {k: v for k, v in client_params.items() if v is not None}

        return client_params


from typing import Any, Callable, Dict, List, Optional
from pydantic import BaseModel, model_validator, field_validator, Field

from gwenflow.utils.tokens import num_tokens_from_string
from gwenflow.types import ChatMessage
from gwenflow.memory.base import BaseChatMemory


DEFAULT_TOKEN_LIMIT = 8192
DEFAULT_TOKEN_LIMIT_RATIO = 0.75


class ChatMemoryBuffer(BaseChatMemory):
 
    token_limit: Optional[int] = Field(None, validate_default=True)
    tokenizer_fn: Optional[Callable] = Field(None, validate_default=True)

    @field_validator("token_limit", mode="before")
    def set_token_limit(cls, v: Optional[int]) -> int:
        token_limit = v or int(DEFAULT_TOKEN_LIMIT * DEFAULT_TOKEN_LIMIT_RATIO)
        return token_limit
    
    @field_validator("tokenizer_fn", mode="before")
    def set_tokenizer_fn(cls, v: Optional[Callable]) -> Callable:
        fn = v or num_tokens_from_string
        return fn
    
    def get(self, initial_token_count: int = 0):

        chat_history = self.get_all()

        if initial_token_count > self.token_limit:
            raise ValueError("Initial token count exceeds token limit")
    
        message_count = len(chat_history)

        cur_messages = chat_history[-message_count:]
        token_count = self._token_count_for_messages(cur_messages) + initial_token_count

        while token_count > self.token_limit and message_count > 1:
            message_count -= 1
            while chat_history[-message_count].role in (
                "tool",
                "assistant",
            ):
                message_count -= 1

            cur_messages = chat_history[-message_count:]
            token_count = (
                self._token_count_for_messages(cur_messages) + initial_token_count
            )

        # catch one message longer than token limit
        if token_count > self.token_limit or message_count <= 0:
            return []

        return chat_history[-message_count:]

    def _token_count_for_messages(self, messages: List[ChatMessage]) -> int:
        if len(messages) <= 0:
            return 0
        text = " ".join(str(m.content) for m in messages)
        return self.tokenizer_fn(text)



import os
import logging
import json

from typing import Optional, Union, Any, List, Dict
from openai import OpenAI, AsyncOpenAI
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk

from gwenflow.types import ChatCompletionMessage, ChatCompletionMessageToolCall
from gwenflow.llms.base import ChatBase


logger = logging.getLogger(__name__)


class ChatOpenAI(ChatBase):
 
    model: str = "gpt-4o-mini"

    # model parameters
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    n: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    max_completion_tokens: Optional[int] = None
    max_tokens: Optional[int] = None
    presence_penalty: Optional[float] = None
    frequency_penalty: Optional[float] = None
    logit_bias: Optional[Dict[int, float]] = None
    response_format: Optional[Dict[str, Any]] = None
    seed: Optional[int] = None
    logprobs: Optional[bool] = None
    top_logprobs: Optional[int] = None

    # clients
    client: Optional[OpenAI] = None
    async_client: Optional[AsyncOpenAI] = None

    # client parameters
    api_key: Optional[str] = None
    organization: Optional[str] = None
    base_url: Optional[str] = None
    timeout: Optional[Union[float, int]] = None
    max_retries: Optional[int] = None

    def _get_client_params(self) -> Dict[str, Any]:

        api_key = self.api_key
        if api_key is None:
            api_key = os.environ.get("OPENAI_API_KEY")
        if api_key is None:
            raise ValueError(
                "The api_key client option must be set either by passing api_key to the client or by setting the OPENAI_API_KEY environment variable"
            )

        organization = self.organization
        if organization is None:
            organization = os.environ.get('OPENAI_ORG_ID')

        client_params = {
            "api_key": api_key,
            "organization": organization,
            "base_url": self.base_url,
            "timeout": self.timeout,
            "max_retries": self.max_retries,
        }

        client_params = {k: v for k, v in client_params.items() if v is not None}

        return client_params

    @property
    def _get_model_params(self) -> Dict[str, Any]:

        model_params = {
            "temperature": self.temperature,
            "top_p": self.top_p,
            "n": self.n,
            "stop": self.stop,
            "max_tokens": self.max_tokens or self.max_completion_tokens,
            "presence_penalty": self.presence_penalty,
            "frequency_penalty": self.frequency_penalty,
            "logit_bias": self.logit_bias,
            "response_format": self.response_format,
            "seed": self.seed,
            "logprobs": self.logprobs,
            "top_logprobs": self.top_logprobs,
        }

        if self.tools:
            tools = [tool.openai_schema for tool in self.tools]
            model_params["tools"] = tools or None
            model_params["tool_choice"] = self.tool_choice or "auto"

        model_params = {k: v for k, v in model_params.items() if v is not None}

        return model_params
    
    def get_client(self) -> OpenAI:

        if self.client:
            return self.client
        
        client_params = self._get_client_params()

        self.client = OpenAI(**client_params)
        return self.client

    def get_async_client(self) -> AsyncOpenAI:

        if self.client:
            return self.client
        
        client_params = self._get_client_params()

        self.async_client = AsyncOpenAI(**client_params)
        return self.async_client

    def _parse_response(self, response, response_format: dict = None):
        """Process the response based on whether tools are used or not."""

        text_response = ""    

        if isinstance(response, ChatCompletionChunk):
            if response.choices[0].delta.content:
                text_response = response.choices[0].delta.content
        else:
            if response.choices[0].message.content:
                text_response = response.choices[0].message.content
        
        if response_format:
            if response_format.get("type") == "json_object":
                text_response = json.loads(text_response)

        return text_response

    def invoke(
        self,
        messages: List[Dict[str, str]],
        parse_response: bool = True,
        **kwargs,
    ):

        response: ChatCompletionMessage = self.get_client().chat.completions.create(
            model=self.model,
            messages=messages,
            **self._get_model_params,
        )

        tool_calls = response.choices[0].message.tool_calls
        if tool_calls and self.tools:
            tool_messages = self.handle_tool_calls(tool_calls=tool_calls)
            if len(tool_messages)>0:
                assistant_message = response.choices[0].message
                messages.append(json.loads(assistant_message.model_dump_json()))
                messages.extend(tool_messages)
                response: ChatCompletionMessage = self.get_client().chat.completions.create(
                    model=self.model,
                    messages=messages,
                    **self._get_model_params,
                )

        if parse_response:
            response = self._parse_response(response, response_format=kwargs.get("response_format"))

        return response

    async def ainvoke(
        self,
        messages: List[Dict[str, str]],
        parse_response: bool = True,
        **kwargs,
    ):

        response: ChatCompletionMessage = await self.get_async_client().chat.completions.create(
            model=self.model,
            messages=messages,
            **self._get_model_params,
        )

        tool_calls = response.choices[0].message.tool_calls
        if tool_calls and self.tools:
            tool_messages = await self.ahandle_tool_calls(tool_calls=tool_calls)
            if len(tool_messages)>0:
                assistant_message = response.choices[0].message
                messages.append(json.loads(assistant_message.model_dump_json()))
                messages.extend(tool_messages)
                response: ChatCompletionMessage = await self.get_async_client().chat.completions.create(
                    model=self.model,
                    messages=messages,
                    **self._get_model_params,
                )

        if parse_response:
            response = self._parse_response(response, response_format=kwargs.get("response_format"))

        return response

    def stream(
        self,
        messages: List[Dict[str, str]],
        parse_response: bool = True,
        **kwargs,
    ):

        streamed_response: ChatCompletionChunk = self.get_client().chat.completions.create(
            model=self.model,
            messages=messages,
            stream=True,
            **self._get_model_params,
        )

        content = ""
        for chunk in streamed_response:
            try:
                if chunk.choices[0].finish_reason == "stop":
                    break
                if chunk.choices[0].delta.content:
                    content += chunk.choices[0].delta.content
                if parse_response:
                    chunk = self._parse_response(chunk, response_format=kwargs.get("response_format"))
                    yield chunk
                else:
                    yield f"data: { json.dumps(chunk.model_dump()) }\n\n"
            
            except Exception as e:
                logger.warning(e)

    async def astream(
        self,
        messages: List[Dict[str, str]],
        parse_response: bool = True,
        **kwargs,
    ):

        response = await self.get_client().chat.completions.create(
            model=self.model,
            messages=messages,
            stream=True,
            **self._get_model_params,
        )

        content = ""
        async for chunk in response:
            if len(chunk.choices) > 0:
                if chunk.choices[0].finish_reason == "stop":
                    break
                if chunk.choices[0].delta.content:
                    content += chunk.choices[0].delta.content
                if parse_response:
                    chunk = self._parse_response(chunk, response_format=kwargs.get("response_format"))
                    yield chunk
                else:
                    yield f"data: { json.dumps(chunk.model_dump()) }\n\n"
 